from converter import (
    artboard,
    group,
    page,
    rectangle,
    shape,
    shape_path,
    shape_group,
    text,
    slice,
    instance,
    symbol,
)
import logging
from sketchformat.layer_common import AbstractLayer
from sketchformat.layer_group import AbstractLayerGroup
from typing import Dict, Callable, Any
import traceback
from .errors import *
from . import utils


def ignored_layer_type(fig_layer: dict) -> AbstractLayer:
    raise Fig2SketchWarning("LAY001")


CONVERTERS: Dict[str, Callable[[dict], AbstractLayer]] = {
    "CANVAS": page.convert,
    "ARTBOARD": artboard.convert,
    "GROUP": group.convert,
    "ROUNDED_RECTANGLE": rectangle.convert,
    "RECTANGLE": rectangle.convert,
    "ELLIPSE": shape.convert_oval,
    "VECTOR": shape_path.convert,
    "STAR": shape.convert_star,
    "REGULAR_POLYGON": shape.convert_polygon,
    "TEXT": text.convert,
    "BOOLEAN_OPERATION": shape_group.convert,
    "LINE": shape_path.convert_line,
    "SLICE": slice.convert,
    "SYMBOL": symbol.convert,
    "INSTANCE": instance.convert,
    "STICKY": ignored_layer_type,
}

POST_PROCESSING: Dict[str, Callable[[dict, Any], AbstractLayer]] = {
    "CANVAS": page.add_page_background,
    "ARTBOARD": artboard.post_process_frame,
    "GROUP": group.post_process_frame,
    "BOOLEAN_OPERATION": shape_group.post_process,
    "SYMBOL": symbol.move_to_symbols_page,
    "INSTANCE": instance.post_process,
}


def convert_node(fig_node: dict, parent_type: str) -> AbstractLayer:
    name = fig_node["name"]
    type_ = get_node_type(fig_node, parent_type)
    logging.debug(f"{type_}: {name} {fig_node['guid']}")

    try:
        sketch_item = CONVERTERS[type_](fig_node)
    except Fig2SketchNodeChanged:
        # The fig_node was modified, retry converting with the new values
        # This happens on instance detaching
        return convert_node(fig_node, parent_type)

    if fig_node.get("layoutGrids", []) and type_ != "ARTBOARD":
        utils.log_conversion_warning("GRD001", fig_node)

    children = []
    for child in fig_node.get("children", []):
        try:
            children.append(convert_node(child, fig_node["type"]))
        except Fig2SketchWarning as w:
            utils.log_conversion_warning(w.code, child)
        except Exception as e:
            logging.error(
                f'An unexpected error occurred when converting {child["type"]}: {child["name"]} ({child["guid"]}). It will be skipped\n'
                + "".join(traceback.format_exception(e))
            )

    if children and isinstance(sketch_item, AbstractLayerGroup):
        sketch_item.layers = children

    post_process = POST_PROCESSING.get(type_)
    if post_process:
        sketch_item = post_process(fig_node, sketch_item)

    return sketch_item


def get_node_type(fig_node: dict, parent_type: str) -> str:
    # We do this because Sketch does not support nested artboards
    # If a Frame is detected inside another Frame, the internal one
    # is considered a group
    if fig_node["type"] in ["FRAME", "SECTION"]:
        if parent_type == "CANVAS" and not fig_node.get("resizeToFit", False):
            return "ARTBOARD"
        else:
            return "GROUP"
    else:
        return fig_node["type"]



from . import style, positioning
from converter import utils
from sketchformat.layer_group import *
from sketchformat.layer_shape import Rectangle
from sketchformat.style import Fill
from typing import Sequence


def convert(fig_canvas: dict) -> Page:
    return make_page(fig_canvas["guid"], fig_canvas["name"])


def symbols_page() -> Page:
    page = make_page((0, 0), "Symbols", suffix=b"symbols_page")

    return page


def make_page(guid: Sequence[int], name: str, suffix: bytes = b"") -> Page:
    return Page(
        do_objectID=utils.gen_object_id(guid, suffix),
        frame=Rect(height=0, width=0, x=0, y=0),
        name=name,
        resizingConstraint=63,
        rotation=0.0,
        style=Style(do_objectID=utils.gen_object_id(guid, suffix + b"style")),
        hasClickThrough=True,
    )


DEFAULT_CANVAS_BACKGROUND = Color(
    red=0.9607843160629272, green=0.9607843160629272, blue=0.9607843160629272, alpha=1
)


def add_page_background(fig_canvas, sketch_page):
    background_color = style.convert_color(
        fig_canvas["backgroundColor"], fig_canvas["backgroundOpacity"]
    )
    if background_color != DEFAULT_CANVAS_BACKGROUND:
        page_bbox = positioning.group_bbox(sketch_page.layers)
        sketch_page.layers.insert(
            0,
            Rectangle(
                do_objectID=utils.gen_object_id(fig_canvas["guid"], b"background"),
                name="Page background",
                style=Style(
                    do_objectID=utils.gen_object_id(fig_canvas["guid"], b"background_style"),
                    fills=[Fill.Color(background_color)],
                ),
                resizingConstraint=0,
                rotation=0,
                frame=Rect(
                    x=page_bbox[0] - 1000,
                    y=page_bbox[2] - 1000,
                    width=(page_bbox[1] - page_bbox[0]) + 2000,
                    height=(page_bbox[3] - page_bbox[2]) + 2000,
                ),
            ),
        )

    return sketch_page



from . import instance, group, base, prototype
from .context import context
from converter import utils
from sketchformat.layer_group import *

LAYOUT_AXIS = {
    "NONE": None,
    "HORIZONTAL": LayoutAxis.HORIZONTAL,
    "VERTICAL": LayoutAxis.VERTICAL,
}

LAYOUT_ANCHOR = {
    "MIN": LayoutAnchor.MIN,
    "CENTER": LayoutAnchor.MIDDLE,
    "MAX": LayoutAnchor.MAX,
    "BASELINE": LayoutAnchor.MIDDLE,
    "SPACE_EVENLY": LayoutAnchor.MIDDLE,
}


def convert(fig_symbol):
    # A symbol is an artboard with a symbolID
    master = SymbolMaster(
        **base.base_styled(fig_symbol),
        **prototype.prototyping_information(fig_symbol),
        symbolID=utils.gen_object_id(fig_symbol["guid"]),
    )

    # Keep the base ID as the symbol reference, create a new one for the container
    master.do_objectID = utils.gen_object_id(fig_symbol["guid"], b"symbol_master")

    # Also add group layout if auto-layout is enabled
    axis = LAYOUT_AXIS[fig_symbol.get("stackMode", "NONE")]
    if axis is not None:
        anchor = LAYOUT_ANCHOR[fig_symbol.get("stackPrimaryAlignItems", "MIN")]

        master.groupLayout = InferredGroupLayout(
            axis=axis,
            layoutAnchor=anchor,
        )

    # Use better names for variants if possible
    try:
        parent = context.fig_node(fig_symbol["parent"]["guid"])
        if parent and parent.get("isStateGroup", False):
            master.name = symbol_variant_name(parent, fig_symbol)

    except Exception as e:
        print(e)

    return master


def move_to_symbols_page(fig_symbol, sketch_symbol):
    if utils.has_rounded_corners(fig_symbol):
        group.create_clip_mask_if_needed(fig_symbol, sketch_symbol)

    # Apply frame transforms
    group.convert_frame_style(fig_symbol, sketch_symbol)

    # After the entire symbol is converted, move it to the Symbols page
    context.add_symbol(sketch_symbol)

    # Since we put the Symbol in a different page in Sketch, leave an instance where the
    # master used to be
    return instance.master_instance(fig_symbol)


def symbol_variant_name(parent, symbol):
    property_values = [x.strip().split("=")[1] for x in symbol["name"].split(",")]
    ordered_values = [parent["name"]] + property_values
    return "/".join(ordered_values)



from .context import context
from .errors import Fig2SketchWarning
from converter import utils
from sketchformat.prototype import *
from typing import TypedDict, Tuple, Optional

OVERLAY_INTERACTION = {
    "NONE": OverlayBackgroundInteraction.NONE,
    "CLOSE_ON_CLICK_OUTSIDE": OverlayBackgroundInteraction.CLOSES_OVERLAY,
}

ANIMATION_TYPE = {
    "INSTANT_TRANSITION": AnimationType.NONE,
    "SLIDE_FROM_LEFT": AnimationType.SLIDE_FROM_LEFT,
    "SLIDE_FROM_RIGHT": AnimationType.SLIDE_FROM_RIGHT,
    "SLIDE_FROM_TOP": AnimationType.SLIDE_FROM_TOP,
    "SLIDE_FROM_BOTTOM": AnimationType.SLIDE_FROM_BOTTOM,
    "PUSH_FROM_LEFT": AnimationType.SLIDE_FROM_LEFT,
    "PUSH_FROM_RIGHT": AnimationType.SLIDE_FROM_RIGHT,
    "PUSH_FROM_TOP": AnimationType.SLIDE_FROM_TOP,
    "PUSH_FROM_BOTTOM": AnimationType.SLIDE_FROM_BOTTOM,
    "MOVE_FROM_LEFT": AnimationType.SLIDE_FROM_LEFT,
    "MOVE_FROM_RIGHT": AnimationType.SLIDE_FROM_RIGHT,
    "MOVE_FROM_TOP": AnimationType.SLIDE_FROM_TOP,
    "MOVE_FROM_BOTTOM": AnimationType.SLIDE_FROM_BOTTOM,
    "SLIDE_OUT_TO_LEFT": AnimationType.SLIDE_FROM_LEFT,
    "SLIDE_OUT_TO_RIGHT": AnimationType.SLIDE_FROM_RIGHT,
    "SLIDE_OUT_TO_TOP": AnimationType.SLIDE_FROM_TOP,
    "SLIDE_OUT_TO_BOTTOM": AnimationType.SLIDE_FROM_BOTTOM,
    "MOVE_OUT_TO_LEFT": AnimationType.SLIDE_FROM_LEFT,
    "MOVE_OUT_TO_RIGHT": AnimationType.SLIDE_FROM_RIGHT,
    "MOVE_OUT_TO_TOP": AnimationType.SLIDE_FROM_TOP,
    "MOVE_OUT_TO_BOTTOM": AnimationType.SLIDE_FROM_BOTTOM,
    "MAGIC_MOVE": AnimationType.NONE,
    "SMART_ANIMATE": AnimationType.NONE,
    "SCROLL_ANIMATE": AnimationType.NONE,
    "DISSOLVE": AnimationType.NONE,
}


class _Flow(TypedDict, total=False):
    flow: FlowConnection


# TODO: Is this called from every node type (groups?)
def convert_flow(fig_node: dict) -> _Flow:
    flow = None
    for interaction in fig_node.get("prototypeInteractions", []):
        if interaction["isDeleted"]:
            continue

        if "event" not in interaction or interaction["event"] == {}:
            continue

        interaction_type = interaction["event"].get("interactionType")
        if interaction_type != "ON_CLICK":
            utils.log_conversion_warning("PRT001", fig_node, props=[interaction_type])
            continue

        for action in interaction["actions"]:
            if flow is not None:
                utils.log_conversion_warning("PRT002", fig_node)
                continue

            # There can be  empty interactions in the model, we just ignore them
            if action == {}:
                continue

            if action.get("navigationType", "NAVIGATE") not in ["NAVIGATE", "SCROLL", "OVERLAY"]:
                utils.log_conversion_warning("PRT003", fig_node, props=[action["navigationType"]])
                continue

            try:
                destination, overlay_settings = get_destination_settings_if_any(action)
            except Fig2SketchWarning as w:
                utils.log_conversion_warning(w.code, fig_node, props=[action["connectionType"]])
                continue

            if destination is not None:
                flow = FlowConnection(
                    destinationArtboardID=destination,
                    animationType=ANIMATION_TYPE[
                        action.get("transitionType", "INSTANT_TRANSITION")
                    ],
                    maintainScrollPosition=action.get("transitionPreserveScroll", False),
                    overlaySettings=overlay_settings,
                )

    return {"flow": flow} if flow else {}


class _PrototypingInformation(TypedDict, total=False):
    isFlowHome: bool
    overlayBackgroundInteraction: OverlayBackgroundInteraction
    presentationStyle: PresentationStyle
    overlaySettings: FlowOverlaySettings
    prototypeViewport: PrototypeViewport


def prototyping_information(fig_frame: dict) -> _PrototypingInformation:
    # Some information about the prototype is in the canvas/page
    fig_canvas = context.fig_node(fig_frame["parent"]["guid"])

    if "prototypeDevice" not in fig_canvas:
        return {
            "isFlowHome": False,
            "overlayBackgroundInteraction": OverlayBackgroundInteraction.NONE,
            "presentationStyle": PresentationStyle.SCREEN,
        }

    # TODO: Overflow scrolling means making the artboard bigger (fit the child bounds)
    if fig_frame.get("scrollDirection", "NONE") != "NONE":
        utils.log_conversion_warning("PRT005", fig_frame)

    if "overlayBackgroundInteraction" in fig_frame:
        return {
            "isFlowHome": False,
            "overlayBackgroundInteraction": OVERLAY_INTERACTION[
                fig_frame["overlayBackgroundInteraction"]
            ],
            "presentationStyle": PresentationStyle.OVERLAY,
            "overlaySettings": FlowOverlaySettings.Positioned(
                fig_frame.get("overlayPositionType", "CENTER")
            ),
        }
    else:
        return {
            "isFlowHome": fig_frame.get("prototypeStartingPoint", {}).get("name", "") != "",
            "prototypeViewport": PrototypeViewport(
                name=fig_canvas["prototypeDevice"]["presetIdentifier"],
                size=Point.from_dict(fig_canvas["prototypeDevice"]["size"]),
            ),
            "overlayBackgroundInteraction": OverlayBackgroundInteraction.NONE,
            "presentationStyle": PresentationStyle.SCREEN,
            "overlaySettings": FlowOverlaySettings.RegularArtboard(),
        }


def get_destination_settings_if_any(
    action: dict,
) -> Tuple[Optional[str], Optional[FlowOverlaySettings]]:
    overlay_settings = None
    destination: Optional[str]

    match action["connectionType"], action.get("transitionNodeID", None):
        case "BACK", _:
            destination = "back"
        case "INTERNAL_NODE", None:
            destination = None
        case "INTERNAL_NODE", transition_node_id:
            if utils.is_invalid_ref(transition_node_id):
                destination = None
            else:
                destination = utils.gen_object_id(transition_node_id)
                transition_node = context.fig_node(transition_node_id)

                if "overlayBackgroundInteraction" in transition_node:
                    offset = action.get("overlayRelativePosition", {"x": 0, "y": 0})

                    overlay_settings = FlowOverlaySettings.Positioned(
                        transition_node.get("overlayPositionType", "CENTER"),
                        Point.from_dict(offset),
                    )

        case "NONE", _:
            destination = None
        case _:
            raise Fig2SketchWarning("PRT004")

    return destination, overlay_settings



import codecs
import ctypes
from collections import OrderedDict
import zlib
import struct
import io
import logging


class KiwiReader:
    def __init__(self, reader):
        self._reader = reader

    def byte(self):
        return self._reader.read(1)[0]

    def bool(self):
        return self.byte() > 0

    def uint(self):
        uint = 0
        shift = 0
        for shift in range(0, 36, 7):
            b = self.byte()
            uint |= (b & 127) << shift
            if b < 128:
                break

        return uint

    def float(self):
        b = self.byte()
        if b == 0:
            return 0.0

        bits = b | self.byte() << 8 | self.byte() << 16 | self.byte() << 24
        bits = (bits << 23) | (bits >> 9)

        return ctypes.c_float.from_buffer(ctypes.c_uint32(bits)).value

    def int(self):
        v = self.uint()
        return ~(v >> 1) if v & 1 else v >> 1

    def string(self):
        string = ""
        decoder = codecs.lookup("utf8").incrementaldecoder()
        while not (string and string[-1] == "\x00"):
            ch = ""
            while not ch:
                ch = decoder.decode(self._reader.read(1))
            string += ch

        return string[:-1]


class KiwiSchema:
    def __init__(self, reader):
        kw = KiwiReader(reader)

        self.types = []
        for _ in range(kw.uint()):
            name = kw.string()
            kind = kw.byte()

            fields = OrderedDict()
            for _ in range(kw.uint()):
                field = KiwiSchema._decode_field(kw)

                fields[field["value"]] = field

            self.types.append({"name": name, "kind": kind, "fields": fields})

    def _decode_field(kw):
        return {
            "name": kw.string(),
            "type": kw.int(),
            "array": kw.bool(),
            "value": kw.uint(),
        }


class KiwiDecoder:
    TYPES = ["bool", "byte", "int", "uint", "float", "string"]
    KINDS = ["ENUM", "STRUCT", "MESSAGE"]

    def __init__(self, schema, type_converters):
        self.schema = schema
        self.type_converters = type_converters

    def decode(self, reader, root):
        kw = KiwiReader(reader)
        root_type = [t for t in self.schema.types if t["name"] == root][0]
        return self._decode_message(kw, root_type)

    def _decode_message(self, kw, type):
        obj = {}
        while (fid := kw.uint()) != 0:
            field = type["fields"][fid]
            ftype = field["type"]

            obj[field["name"]] = self._decode_type(kw, ftype, field["array"])

        return obj

    def _decode_struct(self, kw, type):
        return {
            f["name"]: self._decode_type(kw, f["type"], f["array"])
            for f in type["fields"].values()
        }

    def _decode_enum(self, kw, type):
        value = kw.uint()
        return type["fields"][value]["name"]

    def _decode_type(self, kw, type_id, array):
        obj = self._decode_type_inner(kw, type_id, array)

        type_converter = self.type_converters.get(self.schema.types[type_id]["name"])
        if not array and type_converter:
            obj = type_converter(obj)

        return obj

    def _decode_type_inner(self, kw, type_id, array):
        if array:
            return [self._decode_type(kw, type_id, False) for i in range(kw.uint())]

        if type_id < 0:
            primitive = self.TYPES[~type_id]
            return kw.__getattribute__(primitive)()
        else:
            type = self.schema.types[type_id]
            match type["kind"]:
                case 0:
                    return self._decode_enum(kw, type)
                case 1:
                    return self._decode_struct(kw, type)
                case 2:
                    return self._decode_message(kw, type)
                case other:
                    raise "Unknown"


def decode(reader, type_converters):
    MIN_SUPPORTED_VERSIONS = 15
    MAX_SUPPORTED_VERSION = 25

    header = reader.read(12)
    fig_version = struct.unpack("<I", header[8:12])[0]
    if fig_version < MIN_SUPPORTED_VERSIONS:
        raise Exception(
            f"[FIG002] The .fig file has version {fig_version} which is older than the minimum supported ({MIN_SUPPORTED_VERSIONS}). Cannot convert"
        )
    elif fig_version > MAX_SUPPORTED_VERSION:
        logging.info(
            f"[FIG001] The .fig file has version {fig_version} which is newer than the maximum supported ({MAX_SUPPORTED_VERSION}). Some new properties may not be correctly converted"
        )

    segment_header = reader.read(4)
    size = struct.unpack("<I", segment_header)[0]
    data = io.BytesIO(zlib.decompress(reader.read(size), wbits=-15))
    schema = KiwiSchema(data)

    segment_header = reader.read(4)
    size = struct.unpack("<I", segment_header)[0]
    data = io.BytesIO(zlib.decompress(reader.read(size), wbits=-15))
    return KiwiDecoder(schema, type_converters).decode(data, "Message")



import zipfile
from . import font
from .context import context
from .config import config
from converter import utils
from sketchformat.layer_group import Page
from typing import List
from dataclasses import asdict


def convert(pages: List[Page], output_zip: zipfile.ZipFile) -> dict:
    return {
        "_class": "document",
        "do_objectID": utils.gen_object_id((0, 0), b"document"),
        "assets": {
            "_class": "assetCollection",
            "do_objectID": utils.gen_object_id((0, 0), b"assetCollection"),
            "imageCollection": {"_class": "imageCollection", "images": {}},
            "colorAssets": [],
            "gradientAssets": [],
            "images": [],
            "colors": [],
            "gradients": [],
            "exportPresets": [],
        },
        "colorSpace": context.color_space(),
        "currentPageIndex": 0,
        "foreignLayerStyles": [],
        "foreignSymbols": [],
        "foreignTextStyles": [],
        "foreignSwatches": [],
        "layerStyles": {
            "_class": "sharedStyleContainer",
            "do_objectID": utils.gen_object_id((0, 0), b"sharedStyleContainer"),
            "objects": [],
        },
        "layerSymbols": {
            "_class": "symbolContainer",
            "do_objectID": utils.gen_object_id((0, 0), b"symbolContainer"),
            "objects": [],
        },
        "layerTextStyles": {
            "_class": "sharedTextStyleContainer",
            "do_objectID": utils.gen_object_id((0, 0), b"sharedTextStyleContainer"),
            "objects": [],
        },
        "perDocumentLibraries": [],
        "sharedSwatches": {
            "_class": "swatchContainer",
            "do_objectID": utils.gen_object_id((0, 0), b"swatchContainer"),
            "objects": context.sketch_components(),
        },
        "fontReferences": sorted(
            [
                font.convert(name, font_file, postscript, output_zip)
                for name, (font_file, postscript) in context.used_fonts().items()
                if font_file
            ],
            key=lambda x: x.do_objectID,
        ),
        "documentState": {"_class": "documentState"},
        "pages": [
            {
                "_class": "MSJSONFileReference",
                "_ref_class": "MSImmutablePage",
                "_ref": f"pages/{page.do_objectID}",
            }
            for page in pages
        ],
        "userInfo": {"fig2sketch": {**asdict(config), "salt": config.salt.hex()}},
    }



from enum import IntEnum
from typing import Sequence, TypedDict


class WindingRule(IntEnum):
    NON_ZERO = 0
    EVEN_ODD = 1


class Point:
    class _DictXY(TypedDict, total=False):
        x: float
        y: float

    def __init__(self, x: float, y: float) -> None:
        self.x = x
        self.y = y

    def to_json(self) -> str:
        return f"{{{self.x}, {self.y}}}"

    @staticmethod
    def from_array(array: Sequence[float]) -> "Point":
        return Point(x=array[0], y=array[1])

    @staticmethod
    def from_dict(dict_xy: _DictXY) -> "Point":
        return Point(dict_xy["x"], dict_xy["y"])

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Point):
            return self.x == other.x and self.y == other.y
        return False

    def __add__(self, other: "Point") -> "Point":
        return Point(self.x + other.x, self.y + other.y)

    def __repr__(self):
        return f"Point({self.x}, {self.y})"



from . import base, positioning
import copy
import itertools
from collections import defaultdict
from converter import utils
from sketchformat.layer_group import ShapeGroup, Group
from sketchformat.layer_shape import ShapePath, CurvePoint, CurveMode
from sketchformat.common import WindingRule, Point
from sketchformat.style import MarkerType
from typing import Union, List, TypedDict, Tuple, Dict
from .errors import Fig2SketchWarning

STROKE_CAP_TO_MARKER_TYPE = {
    "NONE": MarkerType.NONE,
    "ARROW_LINES": MarkerType.OPEN_ARROW,
    "ARROW_EQUILATERAL": MarkerType.FILLED_ARROW,
    "TRIANGLE_FILLED": MarkerType.LINE,
    "CIRCLE_FILLED": MarkerType.FILLED_CIRCLE,
    "DIAMOND_FILLED": MarkerType.FILLED_SQUARE,
    "ROUND": MarkerType.NONE,
    "SQUARE": MarkerType.NONE,
}

WINDING_RULE = {"NONZERO": WindingRule.NON_ZERO, "ODD": WindingRule.EVEN_ODD}


class _Points(TypedDict):
    points: List[CurvePoint]
    isClosed: bool


class _Markers(TypedDict, total=False):
    startMarkerType: MarkerType
    endMarkerType: MarkerType


def convert(fig_vector: dict) -> Union[Group, ShapeGroup, ShapePath]:
    if len(fig_vector["vectorNetwork"]["vertices"]) == 0:
        raise Fig2SketchWarning("SHP002")

    fig_regions = get_all_segments(fig_vector["vectorNetwork"])
    regions = [convert_region(fig_vector, region, i) for i, region in enumerate(fig_regions)]

    if len(regions) > 1:
        # Ignore positioning for children. TODO: We should probably be building these shapePaths by
        # hand, instead of relying on the generic convert_shape_path function
        for s in regions:
            s.frame.x = 0
            s.frame.y = 0

        obj = Group(**base.base_styled(fig_vector), layers=regions)  # type: ignore[arg-type]
        obj.style.fills = []

        return obj
    else:
        return regions[0]


def convert_region(
    fig_vector: dict, region: dict, region_index: int = 0
) -> Union[ShapeGroup, ShapePath]:
    loops = [
        convert_shape_path(fig_vector, region["style"], loop, region_index, i)
        for i, loop in enumerate(region["loops"])
    ]

    if len(loops) > 1:
        # Ignore positioning for children.
        for s in loops:
            s.frame.x = 0
            s.frame.y = 0

        obj = ShapeGroup(
            **base.base_styled({**fig_vector, **region["style"]}),
            layers=loops,  # type: ignore[arg-type]
            windingRule=WINDING_RULE[region["windingRule"]],
        )
        obj.do_objectID = utils.gen_object_id(fig_vector["guid"], f"region{region_index}".encode())
        obj.style.do_objectID = utils.gen_object_id(
            fig_vector["guid"], f"style_region{region_index}".encode()
        )

        obj.style.windingRule = obj.windingRule
        return obj
    else:
        return loops[0]


def convert_shape_path(
    fig_vector: dict, style: dict, segments: List[dict], region: int = 0, loop: int = 0
) -> ShapePath:
    points, styles = convert_points(fig_vector, segments)

    obj = ShapePath(
        **base.base_shape({**fig_vector, **style}),
        **points,
    )
    obj.do_objectID = utils.gen_object_id(fig_vector["guid"], f"region{region}loop{loop}".encode())
    obj.style.do_objectID = utils.gen_object_id(
        fig_vector["guid"], f"style_region{region}loop{loop}".encode()
    )

    if styles:
        obj.style.set_markers(styles["startMarkerType"], styles["endMarkerType"])

    return obj


def convert_line(fig_line):
    # Shift line by half its width
    vt = [0, -fig_line["strokeWeight"] / 2, 1]
    vtr = positioning.apply_transform(fig_line, vt)
    fig_line["transform"][0][2] += vtr[0]
    fig_line["transform"][1][2] += vtr[1]

    return ShapePath(
        **base.base_shape(fig_line),
        isClosed=False,
        points=[CurvePoint.Straight(Point(0, 0)), CurvePoint.Straight(Point(1, 1))],
    )


def convert_points(fig_vector: dict, ordered_segments: List[dict]) -> Tuple[_Points, _Markers]:
    vertices = fig_vector["vectorNetwork"]["vertices"]

    is_closed = ordered_segments[0]["start"] == ordered_segments[-1]["end"]

    points_style: _Markers = {}

    if not is_closed:
        first_point = vertices[ordered_segments[0]["start"]]
        last_point = vertices[ordered_segments[-1]["end"]]
        points_style = points_marker_types(fig_vector, first_point, last_point)

    points: Dict[int, CurvePoint] = {}
    for segment in ordered_segments:
        point1, point2 = process_segment(fig_vector, vertices, segment, points)
        points[segment["start"]] = point1
        points[segment["end"]] = point2

    # Reset curve status to straight if no curve handles are set
    # This allows to support corner radius in these cases
    for p in points.values():
        if not p.hasCurveFrom and not p.hasCurveTo:
            p.curveMode = CurveMode.STRAIGHT

    return {"points": list(points.values()), "isClosed": is_closed}, points_style


def get_all_segments(vector_network: dict) -> List[dict]:
    unused_segments = set(range(len(vector_network["segments"])))

    def use_segment(i: int) -> int:
        unused_segments.discard(i)
        return i

    regions = [
        {
            "loops": [
                reorder_segment_points([vector_network["segments"][use_segment(i)] for i in loop])
                for loop in region["loops"]
            ],
            "style": region["style"],
            "windingRule": region["windingRule"],
        }
        for region in vector_network["regions"]
    ]

    if unused_segments:
        loops: List[List[dict]] = reorder_segments(
            [vector_network["segments"][i] for i in unused_segments]
        )

        closed = False
        if len(loops) == 1:
            loop = loops[0]
            if loop[0]["start"] == loop[-1]["end"]:
                closed = True

        rest = {
            "style": {} if closed else {"fillPaints": []},
            "windingRule": "NONZERO",
            "loops": loops,
        }

        regions.append(rest)

    return regions


def swap_segment(segment: dict) -> None:
    segment["start"], segment["end"] = segment["end"], segment["start"]
    segment["tangentStart"], segment["tangentEnd"] = (
        segment["tangentEnd"],
        segment["tangentStart"],
    )


def reorder_segments(segments: List[dict]) -> List[List[dict]]:
    """
    Order segments so that they are continuous

    The input can be in an arbitrary order. This function will try to put the
    segments in order such as seg[n].end == seg[n+1].start.
    """
    # Track which segments are still unordered and start/end on a given point
    segments_with_point = defaultdict(list)
    for s in segments:
        segments_with_point[s["start"]].append(s)
        segments_with_point[s["end"]].append(s)

    total_segments = len(segments)
    count = 0
    runs = []
    while count < total_segments:
        if count == total_segments - 1:
            # Special case when one segment left (e.g: lines)
            # Just return it in the same order as in .fig
            segment = [s[0] for s in segments_with_point.values() if s]
            runs.append(segment)
            break

        ordered_run, run_count = reorder_single_segment(segments_with_point)
        count += run_count
        runs.append(ordered_run)

    return runs


def reorder_single_segment(segments_with_point: Dict[int, List[dict]]) -> Tuple[List[dict], int]:
    # In case the path is open, we try to find an end (a point with a single segment)
    # If we don't, the path should be closed and can choose an arbitrary one by default
    start_segment = [s[0] for s in segments_with_point.values() if s][0]
    for v in segments_with_point.values():
        if len(v) == 1:
            start_segment = v[0]
            break

    # Start with this segment and remove it from consideration
    ordered = [start_segment]
    segments_with_point[start_segment["start"]].remove(start_segment)
    segments_with_point[start_segment["end"]].remove(start_segment)

    # See if we will be able to continue, if not, swap the start segment
    if not segments_with_point[start_segment["end"]]:
        swap_segment(start_segment)

    # Loop until we walked through all the segments or we close the loop
    start_point = ordered[0]["start"]
    count = 1
    while ordered[-1]["end"] != start_point:
        ss = segments_with_point[ordered[-1]["end"]]
        # We should get a single candidate segment for a given point.
        if len(ss) == 0:
            # Cannot continue (open path ends)
            break

        # Remove the segment from future consideration
        segment = ss[0]
        segments_with_point[segment["start"]].remove(segment)
        segments_with_point[segment["end"]].remove(segment)

        # Add the segment to our list, swapping start/end if needed
        if segment["start"] != ordered[-1]["end"]:
            swap_segment(segment)

        ordered.append(segment)
        count += 1

    return ordered, count


def reorder_segment_points(segments: List[dict]) -> List[dict]:
    """
    Make sure segment[0].end == segment[1].start, etc.

    From VectorNetwork docs:
      However, the order of the start and end points in the segments do not matter,
      i.e. the end vertex of one segment does not need to match the start vertex of the next
      segment in the loop, but can instead match the end vertex of that segment.
    """

    # Segments may be reused by other regions in other orders, so let's grab a copy
    segments = copy.deepcopy(segments)

    if len(segments) < 2:
        return segments

    if segments[0]["end"] not in (segments[1]["start"], segments[1]["end"]):
        swap_segment(segments[0])

    for prev, cur in itertools.pairwise(segments):
        if prev["end"] != cur["start"]:
            swap_segment(cur)

    return segments


def process_segment(
    fig_vector: dict, vertices: List[dict], segment: dict, points: Dict[int, CurvePoint]
) -> Tuple[CurvePoint, CurvePoint]:
    point1 = get_or_create_point(fig_vector, points, segment["start"], vertices)
    point2 = get_or_create_point(fig_vector, points, segment["end"], vertices)

    if segment["tangentStart"]["x"] != 0.0 or segment["tangentStart"]["y"] != 0.0:
        vertex1 = vertices[segment["start"]]
        point1.hasCurveFrom = True
        point1.curveFrom = Point.from_dict(vertex1) + Point.from_dict(segment["tangentStart"])
        point1.curveMode = CURVE_MODES[
            vertex1.get("style", {}).get(
                "handleMirroring", fig_vector.get("handleMirroring", "NONE")
            )
        ]

    if segment["tangentEnd"]["x"] != 0.0 or segment["tangentEnd"]["y"] != 0.0:
        vertex2 = vertices[segment["end"]]
        point2.hasCurveTo = True
        point2.curveTo = Point.from_dict(vertex2) + Point.from_dict(segment["tangentEnd"])
        point2.curveMode = CURVE_MODES[
            vertex2.get("style", {}).get(
                "handleMirroring", fig_vector.get("handleMirroring", "NONE")
            )
        ]

    return point1, point2


CURVE_MODES = {
    "STRAIGHT": CurveMode.STRAIGHT,
    "ANGLE_AND_LENGTH": CurveMode.MIRRORED,
    "ANGLE": CurveMode.ASYMMETRIC,
    "NONE": CurveMode.DISCONNECTED,
}


def get_or_create_point(
    fig_vector: dict, points: Dict[int, CurvePoint], index: int, vertices: List[dict]
) -> CurvePoint:
    if index in points:
        point = points[index]
    else:
        fig_point = vertices[index]
        point = CurvePoint.Straight(Point(fig_point["x"], fig_point["y"]))
        point.curveMode = CURVE_MODES[fig_point.get("style", {}).get("handleMirroring", "NONE")]
        point.cornerRadius = fig_point.get("style", {}).get(
            "cornerRadius", fig_vector.get("cornerRadius", 0)
        )

    return point


def points_marker_types(fig_vector: dict, start_point: dict, end_point: dict) -> _Markers:
    start_marker_type = STROKE_CAP_TO_MARKER_TYPE[fig_vector.get("strokeCap", "NONE")]
    end_marker_type = STROKE_CAP_TO_MARKER_TYPE[fig_vector.get("strokeCap", "NONE")]

    if ("style" in start_point) and ("strokeCap" in start_point["style"]):
        start_marker_type = STROKE_CAP_TO_MARKER_TYPE[
            start_point["style"].get("strokeCap", "NONE")
        ]

    if ("style" in end_point) and ("strokeCap" in end_point["style"]):
        end_marker_type = STROKE_CAP_TO_MARKER_TYPE[end_point["style"].get("strokeCap", "NONE")]

    if STROKE_CAP_TO_MARKER_TYPE["TRIANGLE_FILLED"] in [
        start_marker_type,
        end_marker_type,
    ]:
        utils.log_conversion_warning("SHP001", fig_vector)

    return {"startMarkerType": start_marker_type, "endMarkerType": end_marker_type}



import copy
import itertools
from . import base, style
from .context import context
from converter import utils
from sketchformat.text import *
from .font_features import FEATURE_MAPPINGS
from unittest.mock import ANY

AlignVertical = {
    "TOP": TextVerticalAlignment.TOP,
    "CENTER": TextVerticalAlignment.MIDDLE,
    "BOTTOM": TextVerticalAlignment.BOTTOM,
}

AlignHorizontal = {
    "LEFT": TextAlignment.LEFT,
    "CENTER": TextAlignment.CENTER,
    "RIGHT": TextAlignment.RIGHT,
    "JUSTIFIED": TextAlignment.JUSTIFIED,
}

TextCase = {
    "ORIGINAL": TextTransform.NONE,
    "UPPER": TextTransform.UPPERCASE,
    "LOWER": TextTransform.LOWERCASE,
    "TITLE": TextTransform.NONE,
}

TEXT_BEHAVIOUR = {
    "NONE": TextBehaviour.FIXED_WIDTH_AND_HEIGHT,
    "WIDTH_AND_HEIGHT": TextBehaviour.FLEXIBLE_WIDTH,
    "HEIGHT": TextBehaviour.FIXED_WIDTH,
}

# Forces a mask on resizingContraints if the text is set to auto-width/height
CONSTRAINT_MASK_FOR_AUTO_RESIZE = {
    "NONE": 0b111111,
    "WIDTH_AND_HEIGHT": 0b101101,
    "HEIGHT": 0b101111,
}

# Sketch (CoreText?) scales up emojis for small font sizes.
# This table undoes this conversation, so we can match  the .fig doc better
EMOJI_SIZE_ADJUST = {
    2: 1.5,
    3: 2,
    4: 3,
    5: 4,
    6: 5,
    7: 5.5,
    8: 6,
    9: 7,
    10: 7.5,
    11: 8,
    12: 9,
    13: 10,
    14: 10.5,
    15: 11,
    16: 12,
    17: 13,
    18: 13.5,
    19: 14,
    20: 15,
    21: 16,
    22: 18,
    23: 20,
    24: 22,
    25: 24,
}

EMOJI_FONT = "AppleColorEmoji"


def convert(fig_text):
    text_resize = fig_text.get("textAutoResize", "NONE")
    obj = Text(
        **base.base_styled(fig_text),
        attributedString=AttributedString(
            string=fig_text["textData"]["characters"],
            attributes=override_characters_style(fig_text),
        ),
        # No good way to calculate this, so we overestimate by setting the frame
        glyphBounds=Bounds(Point(0, 0), Point.from_dict(fig_text["size"])),
        textBehaviour=TEXT_BEHAVIOUR[text_resize],
    )
    obj.resizingConstraint &= CONSTRAINT_MASK_FOR_AUTO_RESIZE[text_resize]

    obj.style.textStyle = text_style(fig_text)

    # Potentially multiple colors, use the attributes instead of the top level color
    if len(obj.attributedString.attributes) > 1:
        obj.style.fills = []

    # Adjust frame's width to include kerning so the text doesn't get cut
    converted_kerning = obj.style.textStyle.encodedAttributes.kerning
    if obj.textBehaviour == TextBehaviour.FLEXIBLE_WIDTH and converted_kerning is not None:
        obj.frame.width += converted_kerning

    return obj


def text_style(fig_text):
    if fig_text["fontName"]["family"] == EMOJI_FONT:
        font_name = EMOJI_FONT
    else:
        font_name = context.record_font(fig_text["fontName"])

    fills = fig_text.get("fillPaints", [])
    if not fills:
        # Set a transparent fill if no fill is set
        fills = [{"type": "SOLID", "color": {"r": 0, "g": 0, "b": 0, "a": 0}}]

    if len(fills) >= 2:
        utils.log_conversion_warning("TXT002", fig_text)

    if fills[0]["type"] == "SOLID":
        color = style.convert_color(fills[0]["color"])
    else:
        # Text fill is not solid (gradient or image). Sketch doesn't support this
        utils.log_conversion_warning("TXT003", fig_text)
        if fills[0].get("stops", []):
            color = style.convert_color(fills[0]["stops"][0]["color"])
        else:
            color = Color.Black()

    obj = TextStyle(
        encodedAttributes=EncodedAttributes(
            **text_transformation(fig_text),
            MSAttributedStringFontAttribute=FontDescriptor(
                name=font_name, size=fig_text["fontSize"], featureSettings=font_features(fig_text)
            ),
            MSAttributedStringColorAttribute=color,
            textStyleVerticalAlignmentKey=AlignVertical[fig_text.get("textAlignVertical", "TOP")],
            **text_decoration(fig_text),
            kerning=kerning(fig_text),
            paragraphStyle=ParagraphStyle(
                alignment=AlignHorizontal[fig_text.get("textAlignHorizontal", "LEFT")],
                **line_height(fig_text),
            ),
        ),
        verticalAlignment=AlignVertical[fig_text.get("textAlignVertical", "TOP")],
    )

    if fig_text.get("paragraphSpacing", 0) != 0:
        obj.encodedAttributes.paragraphStyle.paragraphSpacing = fig_text["paragraphSpacing"]

    return obj


def override_characters_style(fig_text):
    # The attributes of the Sketch string, our output
    attributes = []

    # Map from styleID to the overridden properties
    override_table = utils.get_style_table_override(fig_text["textData"])

    # List of character styles. For each style, points to the appropriate styleID
    character_styles = fig_text["textData"].get("characterStyleIDs", [])
    all_character_styles = itertools.chain(character_styles, itertools.repeat(0))

    # List of glyphs, taken in pairs (AB, BC, CD). Used to know when to switch from
    # one glyph to another. Used to identify emojis that can span multiple codepoints
    glyphs = fig_text["textData"].get("glyphs")
    if not glyphs:
        # Note, glyphs can be empty when there is a single character and that's still ok
        if len(fig_text["textData"]["characters"]) != 1:
            utils.log_conversion_warning("TXT001", fig_text)

        glyphs = [{"firstCharacter": 0, "styleID": 0}]

    # Add a fake glyph to the end that never gets reached for iteration purposes
    glyph_pairs = itertools.pairwise(glyphs + [{"firstCharacter": -1}])
    current_glyph, next_glyph = next(glyph_pairs)

    # Keep track of what the previous style was and when it started
    last_style = text_style(fig_text)
    first_pos = 0

    # Lengths in .fig docs are given in codepoints. In Sketch, they are given in UTF16 code-units
    # So we keep the Sketch position independently, taking into account UTF16 encoding
    sketch_pos = 0

    # Iterate over all characters in input, including their style id and position
    for pos, (style_id, character) in enumerate(
        zip(all_character_styles, fig_text["textData"]["characters"])
    ):
        # A glyph without a character is an added bullet point (for lists). Skip it
        while "firstCharacter" not in next_glyph:
            utils.log_conversion_warning("TXT005", fig_text)
            current_glyph, next_glyph = next(glyph_pairs)

        # Check if we are still in the same glyph or we have to advance
        # We always advance except in multi-codepoint emojis (e.g: families)
        if pos == next_glyph["firstCharacter"]:
            current_glyph, next_glyph = next(glyph_pairs)

        # Compute the override for this character. Aside from style override,
        # we have to set the emoji font if this is an emoji
        style_override = copy.deepcopy(override_table[style_id])

        is_emoji = "emojiCodePoints" in current_glyph

        if is_emoji:
            style_override["fontName"] = {
                "family": EMOJI_FONT,
                "postscript": EMOJI_FONT,
            }
            # This undoes the tweaking of emoji sizes that Sketch does
            font_size = style_override.get("fontSize", fig_text["fontSize"])
            scaled_font_size = EMOJI_SIZE_ADJUST.get(font_size)
            if scaled_font_size:
                style_override["fontSize"] = scaled_font_size

        # If the style changed (as seen by Sketch), convert the previous style run
        current_style = text_style({**fig_text, **style_override})
        if current_style != last_style and pos != 0:
            attributes.append(
                StringAttribute(
                    location=first_pos,
                    length=sketch_pos - first_pos,
                    attributes=last_style.encodedAttributes,
                )
            )
            first_pos = sketch_pos

        last_style = current_style

        # Characters from supplementary planes are encoded in UTF16 as 2 code units
        # Advance the Sketch position accordingly
        if ord(character) > 0xFFFF:
            sketch_pos += 2
        else:
            sketch_pos += 1

    # Save the last style run
    attributes.append(
        StringAttribute(
            location=first_pos,
            length=sketch_pos - first_pos,
            attributes=last_style.encodedAttributes,
        )
    )

    return attributes


def text_decoration(fig_text):
    decoration = {}

    if "textDecoration" in fig_text:
        match fig_text["textDecoration"]:
            case "UNDERLINE":
                decoration = {"underlineStyle": 1}
            case "STRIKETHROUGH":
                decoration = {"strikethroughStyle": 1}

    return decoration


def kerning(fig_text):
    if "letterSpacing" in fig_text:
        match fig_text["letterSpacing"]:
            case {"units": "PIXELS", "value": pixels}:
                # We handle kerning 0 as Auto (None) to match Sketch behavior better
                return pixels if pixels != 0 else None
            case {"units": "PERCENT", "value": percent}:
                return fig_text["fontSize"] * percent / 100
            case _:
                raise Exception(f"Unknown letter spacing unit")
    else:
        return None


def line_height(fig_text):
    if "lineHeight" in fig_text:
        match fig_text["lineHeight"]["units"]:
            case "PIXELS":
                # Fixed line height
                return {
                    "maximumLineHeight": fig_text["lineHeight"]["value"],
                    "minimumLineHeight": fig_text["lineHeight"]["value"],
                }
            case "PERCENT":
                if fig_text["lineHeight"]["value"] == 100:
                    # Auto (natural baselines)
                    return {}
                else:
                    # Similar to RAW, different natural baseline behaviour. Same comment applies.
                    line_height = round(
                        fig_text["fontSize"] * fig_text["lineHeight"]["value"] / 100
                    )
                    return {
                        "maximumLineHeight": line_height,
                        "minimumLineHeight": line_height,
                    }
            case "RAW":
                # Relative to font size of each line.
                # Sketch does not support this if text sizes change over lines.
                # We just set constant baseline as appropriate for the first line.
                # We can do better using fig.textData.baselines information (applying lineHeight
                # overrides to our attributedString)

                # If < 1, .fig and Sketch calculate the first line position differently
                # Sketch seems to set it to min(lineHeight, lineAscent). We can check
                # baselines[0][position]
                # Maybe we should change the frame position in Sketch to account for this?
                line_height = round(fig_text["fontSize"] * fig_text["lineHeight"]["value"])
                return {
                    "maximumLineHeight": line_height,
                    "minimumLineHeight": line_height,
                }
            case _:
                raise Exception(f"Unknown line height unit")
    else:
        # Default to auto
        return {}


def text_transformation(fig_text):
    if "textCase" in fig_text:
        if fig_text["textCase"] == "TITLE":
            utils.log_conversion_warning("TXT004", fig_text)
        return {"MSAttributedStringTextTransformAttribute": TextCase[fig_text["textCase"]]}
    else:
        return {}


def font_features(fig_text: dict) -> List[OTFeature]:
    sketch_features = []
    unsupported_features = []
    for f in fig_text.get("toggledOnOTFeatures", []):
        try:
            fid, on, off = FEATURE_MAPPINGS[f.lower()]
            sketch_features.append(
                OTFeature(CTFeatureSelectorIdentifier=on, CTFeatureTypeIdentifier=fid)
            )
        except KeyError:
            unsupported_features.append(f)

    for f in fig_text.get("toggledOffOTFeatures", []):
        try:
            fid, on, off = FEATURE_MAPPINGS[f.lower()]
            sketch_features.append(
                OTFeature(CTFeatureSelectorIdentifier=off, CTFeatureTypeIdentifier=fid)
            )
        except KeyError:
            unsupported_features.append(f)

    if unsupported_features:
        utils.log_conversion_warning("TXT006", fig_text, features=unsupported_features)

    return sketch_features



from .layer_common import *
from .common import WindingRule
from .prototype import *
from enum import IntEnum
from typing import Any, Optional, List


class LayoutAxis(IntEnum):
    HORIZONTAL = 0
    VERTICAL = 1


class LayoutAnchor(IntEnum):
    MIN = 0
    MIDDLE = 1
    MAX = 2


@dataclass(kw_only=True)
class RulerData:
    _class: str = field(default="rulerData")
    base: int = 0
    guides: List[int] = field(default_factory=list)


@dataclass(kw_only=True)
class SimpleGrid:
    _class: str = field(default="simpleGrid")
    gridSize: int = 8
    thickGridTimes: int = 1
    isEnabled: bool = False


@dataclass(kw_only=True)
class LayoutGrid:
    _class: str = field(default="layoutGrid")
    columnWidth: int = 0
    gutterHeight: int = 0
    gutterWidth: int = 0
    horizontalOffset: int = 0
    numberOfColumns: int = 0
    rowHeightMultiplication: int = 0
    totalWidth: int = 0
    guttersOutside: bool = False
    drawHorizontal: bool = False
    drawHorizontalLines: bool = False
    drawVertical: bool = False
    isEnabled: bool = True


@dataclass(kw_only=True)
class FreeFormGroupLayout:
    _class: str = field(default="MSImmutableFreeformGroupLayout")


@dataclass(kw_only=True)
class InferredGroupLayout:
    _class: str = field(default="MSImmutableInferredGroupLayout")
    axis: LayoutAxis
    layoutAnchor: LayoutAnchor
    maxSize: int = 0
    minSize: int = 0


@dataclass(kw_only=True)
class AbstractLayerGroup(AbstractStyledLayer):
    hasClickThrough: bool = False
    groupLayout: Union[FreeFormGroupLayout, InferredGroupLayout] = field(
        default_factory=FreeFormGroupLayout
    )
    layers: List[AbstractLayer] = field(default_factory=list)


@dataclass(kw_only=True)
class Page(AbstractLayerGroup):
    _class: str = field(default="page")
    horizontalRulerData: RulerData = field(default_factory=RulerData)
    verticalRulerData: RulerData = field(default_factory=RulerData)
    grid: SimpleGrid = field(default_factory=SimpleGrid)
    layout: Optional[LayoutGrid] = None


@dataclass(kw_only=True)
class ShapeGroup(AbstractLayerGroup):
    _class: str = field(default="shapeGroup")
    windingRule: WindingRule = WindingRule.NON_ZERO  # Legacy, should match style.windingRule


@dataclass(kw_only=True)
class Group(AbstractLayerGroup):
    _class: str = field(default="group")


@dataclass(kw_only=True)
class Artboard(AbstractLayerGroup):
    _class: str = field(default="artboard")
    horizontalRulerData: RulerData = field(default_factory=RulerData)
    verticalRulerData: RulerData = field(default_factory=RulerData)
    grid: Optional[SimpleGrid] = None
    layout: Optional[LayoutGrid] = None
    hasBackgroundColor: bool = False
    backgroundColor: Color = field(default_factory=Color.White)
    includeBackgroundColorInExport: bool = False
    resizesContent: bool = True
    isFlowHome: bool = False
    overlayBackgroundInteraction: OverlayBackgroundInteraction = OverlayBackgroundInteraction.NONE
    presentationStyle: PresentationStyle = PresentationStyle.SCREEN
    overlaySettings: Optional[FlowOverlaySettings] = None
    prototypeViewport: Optional[PrototypeViewport] = None

    shouldBreakMaskChain: bool = True
    layerListExpandedType: LayerListStatus = LayerListStatus.EXPANDED


@dataclass(kw_only=True)
class OverrideProperty:
    _class: str = field(default="overrideProperty")
    overrideName: str
    canOverride: bool


@dataclass(kw_only=True)
class SymbolMaster(Artboard):
    _class: str = field(default="symbolMaster")
    allowsOverrides: bool = True
    includeBackgroundColorInInstance: bool = False
    symbolID: str
    overrideProperties: List[OverrideProperty] = field(default_factory=list)


@dataclass(kw_only=True)
class OverrideValue:
    _class: str = field(default="overrideValue")
    overrideName: str
    value: Any


@dataclass(kw_only=True)
class SymbolInstance(AbstractStyledLayer):
    _class: str = field(default="symbolInstance")
    preservesSpaceWhenHidden: bool = False
    scale: float = 1
    symbolID: str
    overrideValues: List[OverrideValue] = field(default_factory=list)



from . import base
from converter import utils
from sketchformat.layer_common import Rect, ClippingMaskMode
from sketchformat.layer_shape import Rectangle
from sketchformat.style import Style
from typing import Tuple


def convert(fig_rect: dict) -> Rectangle:
    radius, corners = convert_corners(fig_rect)
    return Rectangle(
        **base.base_shape(fig_rect),
        fixedRadius=radius,
        corners=corners,
    )


def convert_corners(fig_rect: dict) -> Tuple[float, Rectangle.Corners]:
    radius = fig_rect.get("cornerRadius", 0)
    fixed = not fig_rect.get("rectangleCornerRadiiIndependent", True)
    return radius, Rectangle.Corners(
        radius if fixed else fig_rect.get("rectangleTopLeftCornerRadius", 0),
        radius if fixed else fig_rect.get("rectangleTopRightCornerRadius", 0),
        radius if fixed else fig_rect.get("rectangleBottomRightCornerRadius", 0),
        radius if fixed else fig_rect.get("rectangleBottomLeftCornerRadius", 0),
    )


def make_clipping_rect(fig: dict, frame: Rect) -> Rectangle:
    obj = make_background_rect(fig, frame, "Clip")
    obj.hasClippingMask = True
    obj.clippingMaskMode = ClippingMaskMode.OUTLINE
    return obj


def make_background_rect(fig: dict, frame: Rect, name: str) -> Rectangle:
    radius, corners = convert_corners(fig)
    return Rectangle(
        do_objectID=utils.gen_object_id(fig["guid"], name.encode()),
        name=name,
        frame=Rect(height=frame.height, width=frame.width, x=0, y=0),
        style=Style(do_objectID=utils.gen_object_id(fig["guid"], f"{name}_style".encode())),
        resizingConstraint=base.HORIZONTAL_CONSTRAINT["STRETCH"]
        + base.VERTICAL_CONSTRAINT["STRETCH"],
        rotation=0,
        fixedRadius=radius,
        corners=corners,
    )



import math
from . import base, group, prototype, rectangle
from converter import utils
from sketchformat.layer_group import Artboard, SimpleGrid, LayoutGrid, Rect
from sketchformat.style import Fill, FillType
from typing import Optional
from collections import namedtuple


def convert(fig_frame: dict) -> Artboard:
    obj = Artboard(
        **base.base_styled(fig_frame),
        **prototype.prototyping_information(fig_frame),
        grid=convert_grid(fig_frame),
    )

    obj.layout = convert_layout(fig_frame, obj.frame)

    return obj


def post_process_frame(fig_frame: dict, sketch_artboard: Artboard) -> Artboard:
    # Sketch only supports one custom color as an artboard background
    # If the frame has more than one color or other custom style we just create
    # the background rectangle with whatever style
    # If the frame/artboard has just one color (and not any other custom style)
    # we set the background color in sketch property
    # We could just always create the rectangle to simplify the logic, but I guess
    # adding always a background rectangle is an overhead for the document itself
    if utils.has_rounded_corners(fig_frame):
        utils.log_conversion_warning("ART001", fig_frame)
        group.create_clip_mask_if_needed(fig_frame, sketch_artboard)

    if sketch_artboard.rotation != 0:
        utils.log_conversion_warning("ART002", fig_frame)

    # The .fig file clips overlays implicitly but .sketch doesn't, so we must add a mask
    if sketch_artboard.overlaySettings is not None:
        sketch_artboard.layers.insert(
            0, rectangle.make_clipping_rect(fig_frame, sketch_artboard.frame)
        )

    match sketch_artboard.style.fills:
        case [Fill(fillType=FillType.COLOR, color=color, isEnabled=True)]:
            # Single color, apply to artboard
            sketch_artboard.backgroundColor = color
            sketch_artboard.hasBackgroundColor = True
            sketch_artboard.style.fills = []

    if sketch_artboard.style.fills or sketch_artboard.style.borders:
        # Anything else, add a background rect
        utils.log_conversion_warning("ART003", fig_frame)
        group.convert_frame_style(fig_frame, sketch_artboard)

    return sketch_artboard


def convert_grid(fig_frame: dict) -> Optional[SimpleGrid]:
    grids = sorted(
        [g for g in fig_frame.get("layoutGrids", []) if g["pattern"] == "GRID"],
        key=lambda x: x["sectionSize"],
    )
    if not grids:
        return None

    primary = grids[0]["sectionSize"]
    secondary = None
    for g in grids[1:]:
        size = g["sectionSize"]
        if size % primary == 0:
            if secondary:
                utils.log_conversion_warning("GRD003", fig_frame)
            else:
                secondary = size
        else:
            utils.log_conversion_warning("GRD002", fig_frame)

    return SimpleGrid(
        gridSize=primary,
        thickGridTimes=secondary / primary if secondary else 0,
        isEnabled=True,
    )


def convert_layout(fig_frame: dict, frame: Rect) -> Optional[LayoutGrid]:
    layouts = [g for g in fig_frame.get("layoutGrids", []) if g["pattern"] == "STRIPES"]

    if not layouts:
        return None

    columns = [l for l in layouts if l["axis"] == "X"]
    if len(columns) > 1:
        utils.log_conversion_warning("GRD004", fig_frame)

    col_config = {}
    if columns:
        sizes = calculate_layout(columns[0], frame.width)

        col_config = {
            "columnWidth": sizes.item_size,
            "gutterWidth": columns[0]["gutterSize"],
            "numberOfColumns": sizes.item_count,
            "totalWidth": sizes.size,
            "drawVertical": True,
            "horizontalOffset": sizes.offset,
        }

    rows = [l for l in layouts if l["axis"] == "Y"]
    if len(rows) > 1:
        utils.log_conversion_warning("GRD004", fig_frame)

    row_config = {}
    if rows:
        gutter_size = rows[0]["gutterSize"]
        sizes = calculate_layout(rows[0], frame.height)

        if sizes.size != frame.height:
            utils.log_conversion_warning("GRD005", fig_frame)

        if sizes.offset != 0:
            utils.log_conversion_warning("GRD006", fig_frame)

        if gutter_size <= 0:
            utils.log_conversion_warning("GRD007", fig_frame)
            gutter_size = 1

        row_scale = sizes.item_size / gutter_size
        int_row_scale = round(row_scale)
        if abs(row_scale - int_row_scale) > 0.01:
            utils.log_conversion_warning("GRD007", fig_frame)
        else:
            row_config = {
                "drawHorizontal": True,
                "gutterHeight": gutter_size,
                "rowHeightMultiplication": int_row_scale,
            }

            if col_config:
                utils.log_conversion_warning("GRD007", fig_frame)
            else:
                row_config["totalWidth"] = frame.width

    if not col_config and not row_config:
        return None

    return LayoutGrid(**col_config, **row_config)


LayoutSizes = namedtuple("LayoutSizes", ["size", "offset", "item_count", "item_size"])


def calculate_layout(layout: dict, size: float) -> LayoutSizes:
    item_num = layout["numSections"]
    gutter_width = layout["gutterSize"]
    item_width = layout["sectionSize"]
    offset = layout["offset"]

    if layout["type"] == "STRETCH":
        if item_num == 2147483647:
            item_num = 1
        total_gutter = (item_num - 1) * gutter_width
        item_width = (size - total_gutter - 2 * offset) / item_num
        if item_width < 0:
            item_width = 0
        layout_size = size
    else:
        if item_num == 2147483647:
            item_num = math.ceil(size / item_width)

        layout_size = item_width * item_num + gutter_width * (item_num - 1)
        if layout["type"] == "MAX":
            offset = size - layout_size
        elif layout["type"] == "CENTER":
            offset = (size - layout_size) / 2

    return LayoutSizes(layout_size, offset, item_num, item_width)



import appdirs
import os
import urllib.request
import urllib.parse
import json
from converter import utils
from fontTools.ttLib import TTFont
from sketchformat.document import FontReference, JsonFileReference
from typing import IO, Tuple
from zipfile import ZipFile

fonts_cache_dir = appdirs.user_cache_dir("Fig2Sketch", "Sketch") + "/fonts"
os.makedirs(fonts_cache_dir, exist_ok=True)


class FontNotFoundError(Exception):
    pass


def retrieve_webfont_family(family):
    WEB_FONT_LIST_URL = "https://fonts.google.com/download/list?family="
    list_url = WEB_FONT_LIST_URL + urllib.parse.quote(family)
    list_response = urllib.request.urlopen(list_url)
    return json.loads(bytearray(list_response.read())[5:])


def download_webfonts(font_file_urls):
    font_files = []

    for fi in font_file_urls:
        filename = fi["filename"].split("/")[-1]
        if not filename.endswith((".ttf", ".otf")):
            continue

        font_file_path = f"{fonts_cache_dir}/{filename}"
        if not os.path.exists(font_file_path):
            urllib.request.urlretrieve(fi["url"], font_file_path)

        font_files.append(font_file_path)

    return font_files


def get_webfont(family, subfamily):
    family_list = retrieve_webfont_family(family)
    font_files = download_webfonts(family_list["manifest"]["fileRefs"])

    for font_file_path in font_files:
        font_file = open(font_file_path, "rb")
        font_names = extract_names(font_file)

        if font_names["family"].lower() == family.lower() and (
            font_names["subfamily"].lower() == subfamily.lower()
            or font_names["subfamily"].replace(" ", "").lower()
            == subfamily.replace(" ", "").lower()
        ):
            font_file.seek(0)
            return font_file, font_names["postscript"]

    raise FontNotFoundError(f"Could not find font {family} {subfamily}")


def convert(
    name: Tuple[str, str], font_file: IO[bytes], postscript: str, output_zip: ZipFile
) -> FontReference:
    family, subfamily = name

    font_file.seek(0)
    data = font_file.read()

    sha = utils.generate_file_ref(data)
    path = f"fonts/{sha}"
    output_zip.open(path, "w").write(data)

    return FontReference(
        do_objectID=utils.gen_object_id((0, 0), bytes.fromhex(sha)),
        fontData=JsonFileReference(_ref_class="MSFontData", _ref=path),
        fontFamilyName=family,
        fontFileName=f"{family}-{subfamily}.ttf",
        postscriptNames=[postscript],
    )


def extract_names(font_file):
    font = TTFont(font_file)
    return {
        "family": font["name"].getBestFamilyName(),
        "subfamily": font["name"].getBestSubFamilyName(),
        "postscript": font["name"].getFirstDebugName((6,)),
    }


def font_matches(fig, font_names):
    return fig["postscript"] == font_names["postscript"] or (
        fig["family"] == font_names["family"] and fig["style"] == font_names["subfamily"]
    )



from sketchformat.layer_group import Page, Point
from typing import List
from . import positioning


# Default area in pixels of the canvas. This is taken from a macbook 16" as a mid-sized device
CANVAS_RESOLUTION = (1200, 900)


def convert(pages: List[Page]) -> dict:
    return {
        "document": {
            "pageListHeight": 200,
            "pageListCollapsed": 0,
            "expandedSymbolPathsInSidebar": [],
            "expandedTextStylePathsInPopover": [],
            "libraryListCollapsed": 0,
        },
        **{page.do_objectID: default_viewport(page) for page in pages},
    }


def default_viewport(page: Page) -> dict:
    if not page.layers:
        return {"scrollOrigin": Point(0, 0), "zoomValue": 1}

    bbox = positioning.group_bbox(page.layers)
    w = bbox[1] - bbox[0]
    h = bbox[3] - bbox[2]

    # Calculate scale to fit bound in expected canvas size, with some margin
    scale = min(CANVAS_RESOLUTION[0] / w, CANVAS_RESOLUTION[1] / h) * 0.8

    # Center the drawing in the expected canvas size
    x = bbox[0] - (CANVAS_RESOLUTION[0] - w * scale) / scale / 2
    y = bbox[2] - (CANVAS_RESOLUTION[1] - h * scale) / scale / 2
    pos = Point(-x * scale, -y * scale)

    return {"scrollOrigin": pos, "zoomValue": scale}



from . import base
from sketchformat.layer_shape import Polygon, Oval, Star


def convert_polygon(fig_polygon):
    return Polygon(**base.base_shape(fig_polygon), numberOfPoints=fig_polygon["count"])


def convert_oval(fig_ellipse):
    return Oval(**base.base_shape(fig_ellipse))


def convert_star(fig_star):
    return Star(
        **base.base_shape(fig_star),
        numberOfPoints=fig_star["count"],
        radius=fig_star["starInnerScale"],
    )



from .common import Point
from .layer_common import AbstractStyledLayer
from .style import Color
from dataclasses import dataclass, field, InitVar
from enum import IntEnum
from typing import Optional, List, Dict


class TextVerticalAlignment(IntEnum):
    TOP = 0
    MIDDLE = 1
    BOTTOM = 2


class TextAlignment(IntEnum):
    LEFT = 0
    RIGHT = 1
    CENTER = 2
    JUSTIFIED = 3


class TextTransform(IntEnum):
    NONE = 0
    UPPERCASE = 1
    LOWERCASE = 2


class UnderlineStyle(IntEnum):
    NONE = 0
    SINGLE = 1


class TextBehaviour(IntEnum):
    FLEXIBLE_WIDTH = 0
    FIXED_WIDTH = 1
    FIXED_WIDTH_AND_HEIGHT = 2


class Bounds:
    def __init__(self, pos: Point, size: Point):
        self.pos = pos
        self.size = size

    def to_json(self):
        return f"{{{self.pos.to_json()}, {self.size.to_json()}}}"


@dataclass(kw_only=True)
class ParagraphStyle:
    _class: str = field(default="paragraphStyle")
    alignment: TextAlignment
    minimumLineHeight: Optional[float] = None
    maximumLineHeight: Optional[float] = None


@dataclass(kw_only=True)
class OTFeature:
    CTFeatureSelectorIdentifier: int
    CTFeatureTypeIdentifier: int


@dataclass(kw_only=True)
class FontDescriptor:
    _class: str = field(default="fontDescriptor")
    name: InitVar[str]
    size: InitVar[float]
    featureSettings: InitVar[List[OTFeature]] = []
    attributes: Dict = field(default_factory=dict)

    def __post_init__(self, name, size, featureSettings):
        self.attributes = {"name": name, "size": size}
        if featureSettings:
            self.attributes["featureSettings"] = featureSettings


@dataclass(kw_only=True)
class EncodedAttributes:
    MSAttributedStringFontAttribute: FontDescriptor
    MSAttributedStringColorAttribute: Color
    textStyleVerticalAlignmentKey: TextVerticalAlignment
    kerning: float
    paragraphStyle: ParagraphStyle
    MSAttributedStringTextTransformAttribute: Optional[TextTransform] = None
    underlineStyle: Optional[UnderlineStyle] = None
    strikethroughStyle: Optional[UnderlineStyle] = None


@dataclass(kw_only=True)
class StringAttribute:
    _class: str = field(default="stringAttribute")
    location: int
    length: int
    attributes: EncodedAttributes


@dataclass(kw_only=True)
class AttributedString:
    _class: str = field(default="attributedString")
    string: str
    attributes: List[StringAttribute]


@dataclass(kw_only=True)
class Text(AbstractStyledLayer):
    _class: str = field(default="text")
    attributedString: AttributedString
    glyphBounds: Bounds
    textBehaviour: TextBehaviour
    automaticallyDrawOnUnderlyingPath: bool = False
    dontSynchroniseWithSymbol: bool = False
    lineSpacingBehaviour: int = 2  # This is more or less a version number


@dataclass(kw_only=True)
class TextStyle:
    _class: str = field(default="textStyle")
    encodedAttributes: EncodedAttributes
    verticalAlignment: TextVerticalAlignment



#
# Copyright © 2017  Google, Inc.
# Copyright © 2018  Ebrahim Byagowi
#
#  This is part of HarfBuzz, a text shaping library.
#
# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the
# above copyright notice and the following two paragraphs appear in
# all copies of this software.
#
# IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE TO ANY PARTY FOR
# DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
# IF THE COPYRIGHT HOLDER HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
#
# THE COPYRIGHT HOLDER SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING,
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
# ON AN "AS IS" BASIS, AND THE COPYRIGHT HOLDER HAS NO OBLIGATION TO
# PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
#
# Google Author(s): Behdad Esfahbod
#

# From https://github.com/harfbuzz/harfbuzz/blob/3d48bfc18731e3c2187a5b0666a7e94dcab0150b/src/hb-aat-layout.h

#
# hb_aat_layout_feature_type_t
#
HB_AAT_LAYOUT_FEATURE_TYPE_ALL_TYPOGRAPHIC = 0
HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES = 1
HB_AAT_LAYOUT_FEATURE_TYPE_CURISVE_CONNECTION = 2
HB_AAT_LAYOUT_FEATURE_TYPE_LETTER_CASE = 3
HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_SUBSTITUTION = 4
HB_AAT_LAYOUT_FEATURE_TYPE_LINGUISTIC_REARRANGEMENT = 5
HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_SPACING = 6
HB_AAT_LAYOUT_FEATURE_TYPE_SMART_SWASH_TYPE = 8
HB_AAT_LAYOUT_FEATURE_TYPE_DIACRITICS_TYPE = 9
HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_POSITION = 10
HB_AAT_LAYOUT_FEATURE_TYPE_FRACTIONS = 11
HB_AAT_LAYOUT_FEATURE_TYPE_OVERLAPPING_CHARACTERS_TYPE = 13
HB_AAT_LAYOUT_FEATURE_TYPE_TYPOGRAPHIC_EXTRAS = 14
HB_AAT_LAYOUT_FEATURE_TYPE_MATHEMATICAL_EXTRAS = 15
HB_AAT_LAYOUT_FEATURE_TYPE_ORNAMENT_SETS_TYPE = 16
HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_ALTERNATIVES = 17
HB_AAT_LAYOUT_FEATURE_TYPE_DESIGN_COMPLEXITY_TYPE = 18
HB_AAT_LAYOUT_FEATURE_TYPE_STYLE_OPTIONS = 19
HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE = 20
HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_CASE = 21
HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING = 22
HB_AAT_LAYOUT_FEATURE_TYPE_TRANSLITERATION = 23
HB_AAT_LAYOUT_FEATURE_TYPE_ANNOTATION_TYPE = 24
HB_AAT_LAYOUT_FEATURE_TYPE_KANA_SPACING_TYPE = 25
HB_AAT_LAYOUT_FEATURE_TYPE_IDEOGRAPHIC_SPACING_TYPE = 26
HB_AAT_LAYOUT_FEATURE_TYPE_UNICODE_DECOMPOSITION_TYPE = 27
HB_AAT_LAYOUT_FEATURE_TYPE_RUBY_KANA = 28
HB_AAT_LAYOUT_FEATURE_TYPE_CJK_SYMBOL_ALTERNATIVES_TYPE = 29
HB_AAT_LAYOUT_FEATURE_TYPE_IDEOGRAPHIC_ALTERNATIVES_TYPE = 30
HB_AAT_LAYOUT_FEATURE_TYPE_CJK_VERTICAL_ROMAN_PLACEMENT_TYPE = 31
HB_AAT_LAYOUT_FEATURE_TYPE_ITALIC_CJK_ROMAN = 32
HB_AAT_LAYOUT_FEATURE_TYPE_CASE_SENSITIVE_LAYOUT = 33
HB_AAT_LAYOUT_FEATURE_TYPE_ALTERNATE_KANA = 34
HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES = 35
HB_AAT_LAYOUT_FEATURE_TYPE_CONTEXTUAL_ALTERNATIVES = 36
HB_AAT_LAYOUT_FEATURE_TYPE_LOWER_CASE = 37
HB_AAT_LAYOUT_FEATURE_TYPE_UPPER_CASE = 38
HB_AAT_LAYOUT_FEATURE_TYPE_LANGUAGE_TAG_TYPE = 39
HB_AAT_LAYOUT_FEATURE_TYPE_CJK_ROMAN_SPACING_TYPE = 103


#
# hb_aat_layout_feature_selector_t
#

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_ALL_TYPOGRAPHIC
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALL_TYPE_FEATURES_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALL_TYPE_FEATURES_OFF = 1

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES
HB_AAT_LAYOUT_FEATURE_SELECTOR_REQUIRED_LIGATURES_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_REQUIRED_LIGATURES_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_COMMON_LIGATURES_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_COMMON_LIGATURES_OFF = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_RARE_LIGATURES_ON = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_RARE_LIGATURES_OFF = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_LOGOS_ON = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_LOGOS_OFF = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_REBUS_PICTURES_ON = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_REBUS_PICTURES_OFF = 9
HB_AAT_LAYOUT_FEATURE_SELECTOR_DIPHTHONG_LIGATURES_ON = 10
HB_AAT_LAYOUT_FEATURE_SELECTOR_DIPHTHONG_LIGATURES_OFF = 11
HB_AAT_LAYOUT_FEATURE_SELECTOR_SQUARED_LIGATURES_ON = 12
HB_AAT_LAYOUT_FEATURE_SELECTOR_SQUARED_LIGATURES_OFF = 13
HB_AAT_LAYOUT_FEATURE_SELECTOR_ABBREV_SQUARED_LIGATURES_ON = 14
HB_AAT_LAYOUT_FEATURE_SELECTOR_ABBREV_SQUARED_LIGATURES_OFF = 15
HB_AAT_LAYOUT_FEATURE_SELECTOR_SYMBOL_LIGATURES_ON = 16
HB_AAT_LAYOUT_FEATURE_SELECTOR_SYMBOL_LIGATURES_OFF = 17
HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_LIGATURES_ON = 18
HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_LIGATURES_OFF = 19
HB_AAT_LAYOUT_FEATURE_SELECTOR_HISTORICAL_LIGATURES_ON = 20
HB_AAT_LAYOUT_FEATURE_SELECTOR_HISTORICAL_LIGATURES_OFF = 21

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES
HB_AAT_LAYOUT_FEATURE_SELECTOR_UNCONNECTED = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_PARTIALLY_CONNECTED = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_CURSIVE = 2

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_LETTER_CASE
HB_AAT_LAYOUT_FEATURE_SELECTOR_UPPER_AND_LOWER_CASE = 0  # deprecated
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALL_CAPS = 1  # deprecated
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALL_LOWER_CASE = 2  # deprecated
HB_AAT_LAYOUT_FEATURE_SELECTOR_SMALL_CAPS = 3  # deprecated
HB_AAT_LAYOUT_FEATURE_SELECTOR_INITIAL_CAPS = 4  # deprecated
HB_AAT_LAYOUT_FEATURE_SELECTOR_INITIAL_CAPS_AND_SMALL_CAPS = 5  # deprecated

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_SUBSTITUTION
HB_AAT_LAYOUT_FEATURE_SELECTOR_SUBSTITUTE_VERTICAL_FORMS_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_SUBSTITUTE_VERTICAL_FORMS_OFF = 1

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_LINGUISTIC_REARRANGEMENT
HB_AAT_LAYOUT_FEATURE_SELECTOR_LINGUISTIC_REARRANGEMENT_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_LINGUISTIC_REARRANGEMENT_OFF = 1

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_SPACING
HB_AAT_LAYOUT_FEATURE_SELECTOR_MONOSPACED_NUMBERS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_NUMBERS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_THIRD_WIDTH_NUMBERS = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_QUARTER_WIDTH_NUMBERS = 3

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_SMART_SWASH_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_WORD_INITIAL_SWASHES_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_WORD_INITIAL_SWASHES_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_WORD_FINAL_SWASHES_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_WORD_FINAL_SWASHES_OFF = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_LINE_INITIAL_SWASHES_ON = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_LINE_INITIAL_SWASHES_OFF = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_LINE_FINAL_SWASHES_ON = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_LINE_FINAL_SWASHES_OFF = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_NON_FINAL_SWASHES_ON = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_NON_FINAL_SWASHES_OFF = 9

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_DIACRITICS_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_SHOW_DIACRITICS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_HIDE_DIACRITICS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_DECOMPOSE_DIACRITICS = 2

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_POSITION
HB_AAT_LAYOUT_FEATURE_SELECTOR_NORMAL_POSITION = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_SUPERIORS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_INFERIORS = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_ORDINALS = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_SCIENTIFIC_INFERIORS = 4

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_FRACTIONS
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_FRACTIONS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_VERTICAL_FRACTIONS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_DIAGONAL_FRACTIONS = 2

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_OVERLAPPING_CHARACTERS_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_PREVENT_OVERLAP_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_PREVENT_OVERLAP_OFF = 1

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_TYPOGRAPHIC_EXTRAS
HB_AAT_LAYOUT_FEATURE_SELECTOR_HYPHENS_TO_EM_DASH_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_HYPHENS_TO_EM_DASH_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_HYPHEN_TO_EN_DASH_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_HYPHEN_TO_EN_DASH_OFF = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_SLASHED_ZERO_ON = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_SLASHED_ZERO_OFF = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_FORM_INTERROBANG_ON = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_FORM_INTERROBANG_OFF = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_SMART_QUOTES_ON = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_SMART_QUOTES_OFF = 9
HB_AAT_LAYOUT_FEATURE_SELECTOR_PERIODS_TO_ELLIPSIS_ON = 10
HB_AAT_LAYOUT_FEATURE_SELECTOR_PERIODS_TO_ELLIPSIS_OFF = 11

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_MATHEMATICAL_EXTRAS
HB_AAT_LAYOUT_FEATURE_SELECTOR_HYPHEN_TO_MINUS_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_HYPHEN_TO_MINUS_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_ASTERISK_TO_MULTIPLY_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_ASTERISK_TO_MULTIPLY_OFF = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_SLASH_TO_DIVIDE_ON = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_SLASH_TO_DIVIDE_OFF = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_INEQUALITY_LIGATURES_ON = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_INEQUALITY_LIGATURES_OFF = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_EXPONENTS_ON = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_EXPONENTS_OFF = 9
HB_AAT_LAYOUT_FEATURE_SELECTOR_MATHEMATICAL_GREEK_ON = 10
HB_AAT_LAYOUT_FEATURE_SELECTOR_MATHEMATICAL_GREEK_OFF = 11

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_ORNAMENT_SETS_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_ORNAMENTS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_DINGBATS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_PI_CHARACTERS = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_FLEURONS = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_DECORATIVE_BORDERS = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_INTERNATIONAL_SYMBOLS = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_MATH_SYMBOLS = 6

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_ALTERNATIVES
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_ALTERNATES = 0

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_DESIGN_COMPLEXITY_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_DESIGN_LEVEL1 = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_DESIGN_LEVEL2 = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_DESIGN_LEVEL3 = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_DESIGN_LEVEL4 = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_DESIGN_LEVEL5 = 4

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_STYLE_OPTIONS
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_STYLE_OPTIONS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_DISPLAY_TEXT = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_ENGRAVED_TEXT = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_ILLUMINATED_CAPS = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_TITLING_CAPS = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_TALL_CAPS = 5

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_CHARACTERS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_SIMPLIFIED_CHARACTERS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS1978_CHARACTERS = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS1983_CHARACTERS = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS1990_CHARACTERS = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_ALT_ONE = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_ALT_TWO = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_ALT_THREE = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_ALT_FOUR = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_ALT_FIVE = 9
HB_AAT_LAYOUT_FEATURE_SELECTOR_EXPERT_CHARACTERS = 10
HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS2004_CHARACTERS = 11
HB_AAT_LAYOUT_FEATURE_SELECTOR_HOJO_CHARACTERS = 12
HB_AAT_LAYOUT_FEATURE_SELECTOR_NLCCHARACTERS = 13
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_NAMES_CHARACTERS = 14

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_CASE
HB_AAT_LAYOUT_FEATURE_SELECTOR_LOWER_CASE_NUMBERS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_UPPER_CASE_NUMBERS = 1

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING
HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_TEXT = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_MONOSPACED_TEXT = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_HALF_WIDTH_TEXT = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_THIRD_WIDTH_TEXT = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_QUARTER_WIDTH_TEXT = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALT_PROPORTIONAL_TEXT = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALT_HALF_WIDTH_TEXT = 6

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_TRANSLITERATION
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_TRANSLITERATION = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_HANJA_TO_HANGUL = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_HIRAGANA_TO_KATAKANA = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_KATAKANA_TO_HIRAGANA = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_KANA_TO_ROMANIZATION = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_ROMANIZATION_TO_HIRAGANA = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_ROMANIZATION_TO_KATAKANA = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_HANJA_TO_HANGUL_ALT_ONE = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_HANJA_TO_HANGUL_ALT_TWO = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_HANJA_TO_HANGUL_ALT_THREE = 9

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_ANNOTATION_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_ANNOTATION = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_BOX_ANNOTATION = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_ROUNDED_BOX_ANNOTATION = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_CIRCLE_ANNOTATION = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_INVERTED_CIRCLE_ANNOTATION = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_PARENTHESIS_ANNOTATION = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_PERIOD_ANNOTATION = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_ROMAN_NUMERAL_ANNOTATION = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_DIAMOND_ANNOTATION = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_INVERTED_BOX_ANNOTATION = 9
HB_AAT_LAYOUT_FEATURE_SELECTOR_INVERTED_ROUNDED_BOX_ANNOTATION = 10

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_KANA_SPACING_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_FULL_WIDTH_KANA = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_KANA = 1

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_IDEOGRAPHIC_SPACING_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_FULL_WIDTH_IDEOGRAPHS = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_IDEOGRAPHS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_HALF_WIDTH_IDEOGRAPHS = 2

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_UNICODE_DECOMPOSITION_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_CANONICAL_COMPOSITION_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_CANONICAL_COMPOSITION_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_COMPATIBILITY_COMPOSITION_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_COMPATIBILITY_COMPOSITION_OFF = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRANSCODING_COMPOSITION_ON = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_TRANSCODING_COMPOSITION_OFF = 5

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_RUBY_KANA
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_RUBY_KANA = (
    0,
)  # deprecated - use HB_AAT_LAYOUT_FEATURE_SELECTOR_RUBY_KANA_OFF instead
HB_AAT_LAYOUT_FEATURE_SELECTOR_RUBY_KANA = (
    1,
)  # deprecated - use HB_AAT_LAYOUT_FEATURE_SELECTOR_RUBY_KANA_ON instead
HB_AAT_LAYOUT_FEATURE_SELECTOR_RUBY_KANA_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_RUBY_KANA_OFF = 3

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_CJK_SYMBOL_ALTERNATIVES_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_CJK_SYMBOL_ALTERNATIVES = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_SYMBOL_ALT_ONE = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_SYMBOL_ALT_TWO = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_SYMBOL_ALT_THREE = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_SYMBOL_ALT_FOUR = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_SYMBOL_ALT_FIVE = 5

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_IDEOGRAPHIC_ALTERNATIVES_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_IDEOGRAPHIC_ALTERNATIVES = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_IDEOGRAPHIC_ALT_ONE = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_IDEOGRAPHIC_ALT_TWO = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_IDEOGRAPHIC_ALT_THREE = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_IDEOGRAPHIC_ALT_FOUR = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_IDEOGRAPHIC_ALT_FIVE = 5

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_CJK_VERTICAL_ROMAN_PLACEMENT_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_VERTICAL_ROMAN_CENTERED = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_VERTICAL_ROMAN_HBASELINE = 1

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_ITALIC_CJK_ROMAN
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_CJK_ITALIC_ROMAN = (
    0,
)  # deprecated - use HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_ITALIC_ROMAN_OFF instead
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_ITALIC_ROMAN = (
    1,
)  # deprecated - use HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_ITALIC_ROMAN_ON instead
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_ITALIC_ROMAN_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_ITALIC_ROMAN_OFF = 3

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_CASE_SENSITIVE_LAYOUT
HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_LAYOUT_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_LAYOUT_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_SPACING_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_SPACING_OFF = 3

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_ALTERNATE_KANA
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_HORIZ_KANA_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_HORIZ_KANA_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_VERT_KANA_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_VERT_KANA_OFF = 3

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES
HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_STYLISTIC_ALTERNATES = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ONE_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ONE_OFF = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWO_ON = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWO_OFF = 5
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THREE_ON = 6
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THREE_OFF = 7
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOUR_ON = 8
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOUR_OFF = 9
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIVE_ON = 10
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIVE_OFF = 11
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIX_ON = 12
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIX_OFF = 13
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVEN_ON = 14
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVEN_OFF = 15
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHT_ON = 16
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHT_OFF = 17
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINE_ON = 18
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINE_OFF = 19
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TEN_ON = 20
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TEN_OFF = 21
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ELEVEN_ON = 22
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ELEVEN_OFF = 23
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWELVE_ON = 24
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWELVE_OFF = 25
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THIRTEEN_ON = 26
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THIRTEEN_OFF = 27
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOURTEEN_ON = 28
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOURTEEN_OFF = 29
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIFTEEN_ON = 30
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIFTEEN_OFF = 31
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIXTEEN_ON = 32
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIXTEEN_OFF = 33
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVENTEEN_ON = 34
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVENTEEN_OFF = 35
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHTEEN_ON = 36
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHTEEN_OFF = 37
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINETEEN_ON = 38
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINETEEN_OFF = 39
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWENTY_ON = 40
HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWENTY_OFF = 41

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_CONTEXTUAL_ALTERNATIVES
HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_ALTERNATES_ON = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_ALTERNATES_OFF = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_SWASH_ALTERNATES_ON = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_SWASH_ALTERNATES_OFF = 3
HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_SWASH_ALTERNATES_ON = 4
HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_SWASH_ALTERNATES_OFF = 5

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_LOWER_CASE
HB_AAT_LAYOUT_FEATURE_SELECTOR_DEFAULT_LOWER_CASE = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_LOWER_CASE_SMALL_CAPS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_LOWER_CASE_PETITE_CAPS = 2

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_UPPER_CASE
HB_AAT_LAYOUT_FEATURE_SELECTOR_DEFAULT_UPPER_CASE = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_UPPER_CASE_SMALL_CAPS = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_UPPER_CASE_PETITE_CAPS = 2

# Selectors for #HB_AAT_LAYOUT_FEATURE_TYPE_CJK_ROMAN_SPACING_TYPE
HB_AAT_LAYOUT_FEATURE_SELECTOR_HALF_WIDTH_CJK_ROMAN = 0
HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_CJK_ROMAN = 1
HB_AAT_LAYOUT_FEATURE_SELECTOR_DEFAULT_CJK_ROMAN = 2
HB_AAT_LAYOUT_FEATURE_SELECTOR_FULL_WIDTH_CJK_ROMAN = 3


# From https://github.com/harfbuzz/harfbuzz/blob/76ce390b5ae15004ff8abf8c6360a377c0656dd6/src/hb-aat-layout.cc
FEATURE_MAPPINGS = {
    "afrc": (
        HB_AAT_LAYOUT_FEATURE_TYPE_FRACTIONS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_VERTICAL_FRACTIONS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_FRACTIONS,
    ),
    "c2pc": (
        HB_AAT_LAYOUT_FEATURE_TYPE_UPPER_CASE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_UPPER_CASE_PETITE_CAPS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_DEFAULT_UPPER_CASE,
    ),
    "c2sc": (
        HB_AAT_LAYOUT_FEATURE_TYPE_UPPER_CASE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_UPPER_CASE_SMALL_CAPS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_DEFAULT_UPPER_CASE,
    ),
    "calt": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CONTEXTUAL_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_ALTERNATES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_ALTERNATES_OFF,
    ),
    "case": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CASE_SENSITIVE_LAYOUT,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_LAYOUT_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_LAYOUT_OFF,
    ),
    "clig": (
        HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_LIGATURES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_LIGATURES_OFF,
    ),
    "cpsp": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CASE_SENSITIVE_LAYOUT,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_SPACING_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CASE_SENSITIVE_SPACING_OFF,
    ),
    "cswh": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CONTEXTUAL_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_SWASH_ALTERNATES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CONTEXTUAL_SWASH_ALTERNATES_OFF,
    ),
    "dlig": (
        HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_RARE_LIGATURES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_RARE_LIGATURES_OFF,
    ),
    "expt": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_EXPERT_CHARACTERS,
        16,
    ),
    "frac": (
        HB_AAT_LAYOUT_FEATURE_TYPE_FRACTIONS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_DIAGONAL_FRACTIONS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_FRACTIONS,
    ),
    "fwid": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_MONOSPACED_TEXT,
        7,
    ),
    "halt": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALT_HALF_WIDTH_TEXT,
        7,
    ),
    "hist": (40, 0, 1),
    "hkna": (
        HB_AAT_LAYOUT_FEATURE_TYPE_ALTERNATE_KANA,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_HORIZ_KANA_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_HORIZ_KANA_OFF,
    ),
    "hlig": (
        HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_HISTORICAL_LIGATURES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_HISTORICAL_LIGATURES_OFF,
    ),
    "hngl": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TRANSLITERATION,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_HANJA_TO_HANGUL,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_TRANSLITERATION,
    ),
    "hojo": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_HOJO_CHARACTERS,
        16,
    ),
    "hwid": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_HALF_WIDTH_TEXT,
        7,
    ),
    "ital": (
        HB_AAT_LAYOUT_FEATURE_TYPE_ITALIC_CJK_ROMAN,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_ITALIC_ROMAN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_CJK_ITALIC_ROMAN_OFF,
    ),
    "jp04": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS2004_CHARACTERS,
        16,
    ),
    "jp78": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS1978_CHARACTERS,
        16,
    ),
    "jp83": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS1983_CHARACTERS,
        16,
    ),
    "jp90": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_JIS1990_CHARACTERS,
        16,
    ),
    "liga": (
        HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_COMMON_LIGATURES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_COMMON_LIGATURES_OFF,
    ),
    "lnum": (
        HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_CASE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_UPPER_CASE_NUMBERS,
        2,
    ),
    "mgrk": (
        HB_AAT_LAYOUT_FEATURE_TYPE_MATHEMATICAL_EXTRAS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_MATHEMATICAL_GREEK_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_MATHEMATICAL_GREEK_OFF,
    ),
    "nlck": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NLCCHARACTERS,
        16,
    ),
    "onum": (
        HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_CASE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_LOWER_CASE_NUMBERS,
        2,
    ),
    "ordn": (
        HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_POSITION,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ORDINALS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NORMAL_POSITION,
    ),
    "palt": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALT_PROPORTIONAL_TEXT,
        7,
    ),
    "pcap": (
        HB_AAT_LAYOUT_FEATURE_TYPE_LOWER_CASE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_LOWER_CASE_PETITE_CAPS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_DEFAULT_LOWER_CASE,
    ),
    "pkna": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_TEXT,
        7,
    ),
    "pnum": (
        HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_NUMBERS,
        4,
    ),
    "pwid": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_PROPORTIONAL_TEXT,
        7,
    ),
    "qwid": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_QUARTER_WIDTH_TEXT,
        7,
    ),
    "rlig": (
        HB_AAT_LAYOUT_FEATURE_TYPE_LIGATURES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_REQUIRED_LIGATURES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_REQUIRED_LIGATURES_OFF,
    ),
    "ruby": (
        HB_AAT_LAYOUT_FEATURE_TYPE_RUBY_KANA,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_RUBY_KANA_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_RUBY_KANA_OFF,
    ),
    "sinf": (
        HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_POSITION,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SCIENTIFIC_INFERIORS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NORMAL_POSITION,
    ),
    "smcp": (
        HB_AAT_LAYOUT_FEATURE_TYPE_LOWER_CASE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_LOWER_CASE_SMALL_CAPS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_DEFAULT_LOWER_CASE,
    ),
    "smpl": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SIMPLIFIED_CHARACTERS,
        16,
    ),
    "ss01": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ONE_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ONE_OFF,
    ),
    "ss02": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWO_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWO_OFF,
    ),
    "ss03": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THREE_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THREE_OFF,
    ),
    "ss04": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOUR_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOUR_OFF,
    ),
    "ss05": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIVE_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIVE_OFF,
    ),
    "ss06": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIX_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIX_OFF,
    ),
    "ss07": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVEN_OFF,
    ),
    "ss08": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHT_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHT_OFF,
    ),
    "ss09": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINE_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINE_OFF,
    ),
    "ss10": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TEN_OFF,
    ),
    "ss11": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ELEVEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_ELEVEN_OFF,
    ),
    "ss12": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWELVE_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWELVE_OFF,
    ),
    "ss13": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THIRTEEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_THIRTEEN_OFF,
    ),
    "ss14": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOURTEEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FOURTEEN_OFF,
    ),
    "ss15": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIFTEEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_FIFTEEN_OFF,
    ),
    "ss16": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIXTEEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SIXTEEN_OFF,
    ),
    "ss17": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVENTEEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_SEVENTEEN_OFF,
    ),
    "ss18": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHTEEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_EIGHTEEN_OFF,
    ),
    "ss19": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINETEEN_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_NINETEEN_OFF,
    ),
    "ss20": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLISTIC_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWENTY_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_STYLISTIC_ALT_TWENTY_OFF,
    ),
    "subs": (
        HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_POSITION,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_INFERIORS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NORMAL_POSITION,
    ),
    "sups": (
        HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_POSITION,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SUPERIORS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NORMAL_POSITION,
    ),
    "swsh": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CONTEXTUAL_ALTERNATIVES,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SWASH_ALTERNATES_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SWASH_ALTERNATES_OFF,
    ),
    "titl": (
        HB_AAT_LAYOUT_FEATURE_TYPE_STYLE_OPTIONS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_TITLING_CAPS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_NO_STYLE_OPTIONS,
    ),
    "tnam": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_NAMES_CHARACTERS,
        16,
    ),
    "tnum": (
        HB_AAT_LAYOUT_FEATURE_TYPE_NUMBER_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_MONOSPACED_NUMBERS,
        4,
    ),
    "trad": (
        HB_AAT_LAYOUT_FEATURE_TYPE_CHARACTER_SHAPE,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_TRADITIONAL_CHARACTERS,
        16,
    ),
    "twid": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_THIRD_WIDTH_TEXT,
        7,
    ),
    "unic": (HB_AAT_LAYOUT_FEATURE_TYPE_LETTER_CASE, 14, 15),
    "valt": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALT_PROPORTIONAL_TEXT,
        7,
    ),
    "vert": (
        HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_SUBSTITUTION,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SUBSTITUTE_VERTICAL_FORMS_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SUBSTITUTE_VERTICAL_FORMS_OFF,
    ),
    "vhal": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALT_HALF_WIDTH_TEXT,
        7,
    ),
    "vkna": (
        HB_AAT_LAYOUT_FEATURE_TYPE_ALTERNATE_KANA,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_VERT_KANA_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALTERNATE_VERT_KANA_OFF,
    ),
    "vpal": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TEXT_SPACING,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_ALT_PROPORTIONAL_TEXT,
        7,
    ),
    "vrt2": (
        HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_SUBSTITUTION,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SUBSTITUTE_VERTICAL_FORMS_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SUBSTITUTE_VERTICAL_FORMS_OFF,
    ),
    "vrtr": (HB_AAT_LAYOUT_FEATURE_TYPE_VERTICAL_SUBSTITUTION, 2, 3),
    "zero": (
        HB_AAT_LAYOUT_FEATURE_TYPE_TYPOGRAPHIC_EXTRAS,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SLASHED_ZERO_ON,
        HB_AAT_LAYOUT_FEATURE_SELECTOR_SLASHED_ZERO_OFF,
    ),
}



from converter import utils
from . import base, positioning, rectangle
from sketchformat.layer_group import (
    Group,
    Rect,
    AbstractStyledLayer,
    AbstractLayerGroup,
)
from sketchformat.style import *


def convert(fig_group):
    return Group(
        **base.base_styled(fig_group),
    )


def post_process_frame(fig_group: dict, sketch_group: Group) -> Group:
    convert_frame_style(fig_group, sketch_group)

    if fig_group.get("resizeToFit", False):
        adjust_group_resizing_constraint(fig_group, sketch_group)
    else:
        # For frames converted to groups, add clipmask, resize children, etc
        convert_frame_to_group(fig_group, sketch_group)

    return sketch_group


def adjust_group_resizing_constraint(fig_group: dict, sketch_group: Group) -> None:
    """Adjust the resizing constraint of the group to better match the .fig doc.

    Groups in .fig don't really have a resizing constraint. Instead, the children of the group resize
    relative to the parent of the group.

    If all childs have the same constraint, we can have the same behaviour in Sketch, by setting the group
    constraints to be equal to the sublayers.
    However, if there is a mix, we cannot replicate the behaviour, so we just choose some constraint and throw
    a warning"""
    if not sketch_group.layers:
        return

    constraint = sketch_group.layers[0].resizingConstraint
    if any([l.resizingConstraint != constraint for l in sketch_group.layers[1:]]):
        utils.log_conversion_warning("GRP002", fig_group)

    sketch_group.resizingConstraint = constraint


def convert_frame_to_group(fig_group: dict, sketch_group: AbstractLayerGroup) -> None:
    has_clip_mask = create_clip_mask_if_needed(fig_group, sketch_group)

    if not has_clip_mask:
        # When converting from a frame to a group, the bounding box should be adjusted
        # The frame box in a fig doc can be smalled than the children bounds, but not so in Sketch
        # To do so, we resize the frame to match the children bbox and also move the children
        # so that the top-left corner sits at 0,0
        children_bbox = positioning.group_bbox(sketch_group.layers)
        vector = positioning.Vector(children_bbox[0], children_bbox[2])

        # Translate children
        for child in sketch_group.layers:
            child.frame.x -= vector[0]
            child.frame.y -= vector[1]

        # Translate group
        tr_vector = positioning.apply_transform(fig_group, vector)
        w = children_bbox[1] - children_bbox[0]
        h = children_bbox[3] - children_bbox[2]
        new_xy = positioning.transform_frame(fig_group, {"x": w, "y": h}) + tr_vector

        sketch_group.frame = Rect(x=new_xy[0], y=new_xy[1], width=w, height=h)


def create_clip_mask_if_needed(fig_group: dict, sketch_group: AbstractLayerGroup) -> bool:
    needs_clip_mask = not fig_group.get("frameMaskDisabled", False)
    if needs_clip_mask:
        # Add a clipping rectangle matching the frame size. No need to recalculate bounds
        # since the clipmask defines Sketch bounds (which match visible children)
        sketch_group.layers.insert(0, rectangle.make_clipping_rect(fig_group, sketch_group.frame))

    return needs_clip_mask


def convert_frame_style(fig_group: dict, sketch_group: AbstractLayerGroup) -> AbstractLayerGroup:
    # Convert frame styles
    # - Fill/stroke/bgblur -> Rectangle on bottom with that style
    # - Layer blur -> Rectangle with bgblur on top
    # - Shadows -> Keep in the group
    # TODO: This is one case where we could have both background blur and layer blur
    style = sketch_group.style

    # Fill and borders go on a rectangle on the bottom
    has_fills = any([f.isEnabled for f in style.fills])
    has_borders = any([b.isEnabled for b in style.borders])
    has_inner_shadows = any([b.isEnabled for b in style.innerShadows])
    has_bgblur = style.blur.isEnabled and style.blur.type == BlurType.BACKGROUND
    has_blur = style.blur.isEnabled and style.blur.type == BlurType.GAUSSIAN

    if has_fills or has_borders or has_bgblur:
        bgrect = rectangle.make_background_rect(fig_group, sketch_group.frame, "Frame Background")
        bgrect.style.fills = style.fills
        bgrect.style.borders = style.borders
        if has_bgblur:
            bgrect.style.blur = Blur(type=BlurType.BACKGROUND, radius=style.blur.radius)
        bgrect.style.innerShadows = style.innerShadows

        sketch_group.layers.insert(0, bgrect)

        style.fills = []
        style.borders = []
        style.blur.isEnabled = False
        style.innerShadows = []

    # Blur goes in a rectangle with bgblur at the top
    if has_blur:
        blur = rectangle.make_background_rect(
            fig_group, sketch_group.frame, f"{sketch_group.name} blur"
        )
        blur.style.blur = Blur(type=BlurType.BACKGROUND, radius=style.blur.radius)

        # Foreground blur, add as a layer at the top of the group
        sketch_group.layers.append(blur)
        style.blur.isEnabled = False

    # Inner shadows apply to each child (if they were not put in the background rect earlier)
    # Normal shadows are untouched
    for shadow in style.innerShadows:
        utils.log_conversion_warning("GRP001", fig_group)
        apply_inner_shadow(sketch_group, shadow)

    style.innerShadows = []

    return sketch_group


def apply_inner_shadow(layer: AbstractLayerGroup, shadow: InnerShadow) -> None:
    for child in layer.layers:
        if isinstance(child, AbstractLayerGroup):
            apply_inner_shadow(child, shadow)
        elif isinstance(child, AbstractStyledLayer):
            child.style.innerShadows.append(shadow)



import io
import logging
import shutil
from typing import Tuple, Sequence, Dict
from converter import utils
from zipfile import ZipFile
from . import decodefig, vector_network
from PIL import Image, UnidentifiedImageError, ImageFile


def convert_fig(path: str, output: ZipFile) -> Tuple[dict, Dict[Sequence[int], dict]]:
    fig, fig_zip = decodefig.decode(path)  # type: ignore [no-untyped-call]

    if fig_zip and output:
        shutil.copyfileobj(
            fig_zip.open(f"thumbnail.png", "r"),
            output.open(f"previews/preview.png", "w"),
        )

    # Load all nodes into a map
    id_map = {}
    override_map = {}
    root = None

    for node in fig["nodeChanges"]:
        node = transform_node(fig, node, fig_zip, output)  # type: ignore [no-untyped-call]
        node_id = node["guid"]
        id_map[node_id] = node

        if "overrideKey" in node:
            if node["overrideKey"][0] == utils.UINT_MAX:
                del node["overrideKey"]
            else:
                override_map[node["overrideKey"]] = node

        if not root:
            root = node_id

    # Build the tree
    tree = {"document": id_map[root]}
    for node in id_map.values():
        if "parent" not in node:
            continue

        id_map[node["parent"]["guid"]]["children"].append(node)

    # Sort children
    for node in id_map.values():
        node["children"].sort(key=lambda n: n["parent"]["position"])

    id_map.update(override_map)

    return tree, id_map


def transform_node(fig, node, fig_zip, output):
    node["children"] = []

    # Extract parent ID
    if "parentIndex" in node:
        parent = node.pop("parentIndex")
        node["parent"] = {"guid": parent["guid"], "position": parent["position"]}

    if "vectorData" in node:
        blob_id = node["vectorData"]["vectorNetworkBlob"]
        scale = node["vectorData"]["normalizedSize"]
        style_table_override = utils.get_style_table_override(node["vectorData"])
        network = vector_network.decode(fig, blob_id, scale, style_table_override)
        node["vectorNetwork"] = network

    # Images
    for paint in node.get("fillPaints", []):
        if "image" in paint:
            fname = bytes(paint["image"]["hash"]).hex()
            blob_id = paint["image"].get("dataBlob")
            blob = bytes(fig["blobs"][blob_id]["bytes"]) if blob_id else None
            paint["image"]["filename"] = convert_image(fname, blob, fig_zip, output)

    if "symbolData" in node:
        for override in node["symbolData"].get("symbolOverrides", []):
            for paint in override.get("fillPaints", []):
                if "image" in paint:
                    fname = bytes(paint["image"]["hash"]).hex()
                    blob_id = paint["image"].get("dataBlob")
                    blob = bytes(fig["blobs"][blob_id]["bytes"]) if blob_id else None
                    paint["image"]["filename"] = convert_image(fname, blob, fig_zip, output)

    return node


converted_images: Dict[str, str] = {}


def convert_image(fname, blob, fig_zip, output):
    img = converted_images.get(fname)
    if img:
        return img

    logging.debug(f"Converting image {fname}")
    try:
        if fig_zip is not None:
            fd = fig_zip.open(f"images/{fname}")
        else:
            fd = io.BytesIO(blob)

        image = Image.open(fd)

        # Save to memory, calculate hash, and save
        out = io.BytesIO()
        if image.format in ["PNG", "JPEG"]:
            # No need to convert if it's already a PNG or JPEG
            fd.seek(0)
            while buf := fd.read(16536):
                out.write(buf)
            extension = "" if image.format == "JPEG" else ".png"
        else:
            image.save(out, format="png")
            extension = ".png"

        fhash = utils.generate_file_ref(out.getbuffer())
        output.open(f"images/{fhash}{extension}", "w").write(out.getbuffer())
        converted_images[fname] = f"{fhash}{extension}"
        return f"{fhash}{extension}"
    except UnidentifiedImageError as e:
        if ImageFile.LOAD_TRUNCATED_IMAGES:
            utils.log_conversion_warning("IMG002", {"type": "Image", "guid": fname, "name": fname})
        else:
            utils.log_conversion_warning("IMG001", {"type": "Image", "guid": fname, "name": fname})

        return "f2s_corrupted"
    except KeyError:
        utils.log_conversion_warning("IMG003", {"type": "Image", "guid": fname, "name": fname})
        return "f2s_missing"
    except Exception as e:
        utils.log_conversion_warning(
            "IMG004", {"type": "Image", "guid": fname, "name": fname}, error=str(e)
        )

        return "f2s_corrupted"
